# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import ipaddress
from textwrap import dedent
from typing import Optional, List, Dict

import os

from kubemarine import plugins, kubernetes
from kubemarine.core import utils, log
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.kubernetes import secrets
from kubemarine.plugins.manifest import Processor, EnrichmentFunction, Manifest, Identity


def enrich_inventory(inventory: dict, cluster: KubernetesCluster) -> dict:
    if not inventory["plugins"]["calico"]["install"]:
        return inventory

    # if user defined resources himself, we should use them as is, instead of merging with our defaults
    raw_calico_node = cluster.raw_inventory.get("plugins", {}).get("calico", {}).get("node", {})
    if "resources" in raw_calico_node:
        inventory["plugins"]["calico"]["node"]["resources"] = raw_calico_node["resources"]
    raw_typha = cluster.raw_inventory.get("plugins", {}).get("calico", {}).get("typha", {})
    if "resources" in raw_typha:
        inventory["plugins"]["calico"]["typha"]["resources"] = raw_typha["resources"]
    raw_calico_controller = cluster.raw_inventory.get("plugins", {}).get("calico", {}).get("kube-controllers", {})
    if "resources" in raw_calico_controller:
        inventory["plugins"]["calico"]["kube-controllers"]["resources"] = raw_calico_controller["resources"]
    raw_apiserver = cluster.raw_inventory.get("plugins", {}).get("calico", {}).get("apiserver", {})
    if "resources" in raw_apiserver:
        inventory["plugins"]["calico"]["apiserver"]["resources"] = raw_apiserver["resources"]

    return inventory


# DEPRECATED
def apply_calico_yaml(cluster: KubernetesCluster, calico_original_yaml: str, calico_yaml: str) -> None:
    """
    The method implements full processing for Calico plugin
    :param calico_original_yaml: path to original Calico manifest
    :param calico_yaml: file name of the resulting Calico manifest
    :param cluster: Cluster object
    """

    calico_yaml = os.path.basename(calico_yaml)
    processor = CalicoManifestProcessor(cluster.log, cluster.inventory,
                                        original_yaml_path=calico_original_yaml,
                                        destination_name=calico_yaml)
    manifest = processor.enrich()
    processor.apply(cluster, manifest)


def is_typha_enabled(inventory: dict) -> bool:
    str_value = utils.true_or_false(inventory['plugins']['calico']['typha']['enabled'])
    if str_value == 'false':
        return False
    elif str_value == 'true':
        return True
    else:
        raise Exception(f"plugins.calico.typha.enabled must be set in 'True' or 'False' "
                        f"as string or boolean value")


def is_apiserver_enabled(inventory: dict) -> bool:
    enabled: bool = inventory['plugins']['calico']['apiserver']['enabled']
    return enabled


def renew_apiserver_certificate(cluster: KubernetesCluster) -> None:
    logger = cluster.log
    if not is_apiserver_enabled(cluster.inventory):
        logger.debug("Calico API server is disabled. Skip renewing of the key and certificate.")
        return

    namespace = "calico-apiserver"
    secret_name = "calico-apiserver-certs"
    deployment = "calico-apiserver"

    control_planes = cluster.nodes["control-plane"]
    first_control_plane = control_planes.get_first_member()

    with secrets.create_tls_secret_procedure(first_control_plane):
        logger.debug("Creating or renewing of the key and certificate for the Calico API server")

        config = dedent(
            """\
            [req]
            distinguished_name = req
            [v3_req]
            basicConstraints = critical,CA:TRUE
            subjectAltName = DNS:calico-api.calico-apiserver.svc
            """
        )
        first_control_plane.call(
            secrets.create_certificate,
            config=config,
            customization_flags='-newkey rsa:4096 -days 365 -subj "/" -extensions v3_req')

        first_control_plane.call(secrets.renew_tls_secret, name=secret_name, namespace=namespace)

        logger.debug("Patching the APIService for the Calico API server")

        first_control_plane.sudo(
            f"{secrets.get_encoded_certificate_cmd()} "
            "| xargs -I CERT "
            "sudo kubectl patch apiservice v3.projectcalico.org "
            r'-p "{\"spec\": {\"caBundle\": \"CERT\"}}"')

    # Force restart the deployment instead of graceful waiting for the automatic secret propagation.
    # This is necessary to make the procedure independent on the secret propagation timeout.
    logger.debug("Restarting the Calico API server deployment")
    first_control_plane.sudo(f"kubectl rollout restart -n {namespace} deployment {deployment}")
    plugins.expect_deployment(cluster, [{'name': deployment, 'namespace': namespace}])
    plugins.expect_pods(cluster, ['calico-apiserver'], namespace=namespace)

    logger.debug("Waiting for the Calico API service availability through the Kubernetes API server")

    # Try to access some projectcalico.org resource using each instance of the Kubernetes API server.
    expect_config = cluster.inventory['plugins']['calico']['apiserver']['expect']['apiservice']
    with kubernetes.local_admin_config(control_planes) as kubeconfig:
        control_planes.call(utils.wait_command_successful,
                            command=f"kubectl --kubeconfig {kubeconfig} get ippools.projectcalico.org",
                            hide=True,
                            retries=expect_config['retries'], timeout=expect_config['timeout'])


class CalicoManifestProcessor(Processor):
    def __init__(self, logger: log.VerboseLogger, inventory: dict,
                 original_yaml_path: Optional[str] = None, destination_name: Optional[str] = None):
        super().__init__(logger, inventory, Identity('calico'), original_yaml_path, destination_name)

    def exclude_typha_objects_if_disabled(self, manifest: Manifest) -> None:
        # enrich 'calico-typha' objects only if it's enabled in 'cluster.yaml'
        # in other case those objects must be excluded
        if not is_typha_enabled(self.inventory):
            for key in ("Deployment_calico-typha", "Service_calico-typha", "PodDisruptionBudget_calico-typha"):
                self.exclude(manifest, key)

    def enrich_configmap_calico_config(self, manifest: Manifest) -> None:
        """
        The method implements the enrichment procedure for Calico ConfigMap
        :param manifest: Container to operate with manifest objects
        """

        key = "ConfigMap_calico-config"
        source_yaml = manifest.get_obj(key, patch=True)
        val = self.inventory['plugins']['calico']['mtu']
        source_yaml['data']['veth_mtu'] = str(val)
        self.log.verbose(f"The {key} has been patched in 'data.veth_mtu' with '{val}'")
        val = "calico-typha" if is_typha_enabled(self.inventory) else "none"
        source_yaml['data']['typha_service_name'] = val
        self.log.verbose(f"The {key} has been patched in 'data.typha_service_name' with '{val}'")
        string_part = source_yaml['data']['cni_network_config']
        ip = self.inventory['services']['kubeadm']['networking']['podSubnet'].split('/')[0]
        if type(ipaddress.ip_address(ip)) is ipaddress.IPv4Address:
            val = self.inventory['plugins']['calico']['cni']['ipam']['ipv4']
        else:
            val = self.inventory['plugins']['calico']['cni']['ipam']['ipv6']
        new_string_part = string_part.replace('"type": "calico-ipam"', str(val)[:-1][1:].replace("'", "\""))
        source_yaml['data']['cni_network_config'] = new_string_part
        log_str = new_string_part.replace("\n", "")
        self.log.verbose(f"The {key} has been patched in 'data.cni_network_config' with '{log_str}'")

    def enrich_deployment_calico_kube_controllers(self, manifest: Manifest) -> None:
        """
        The method implements the enrichment procedure for Calico controller Deployment
        :param manifest: Container to operate with manifest objects
        """

        key = "Deployment_calico-kube-controllers"
        self.enrich_node_selector(manifest, key, plugin_service='kube-controllers')
        self.enrich_tolerations(manifest, key, plugin_service='kube-controllers')
        self.enrich_resources_for_container(manifest, key,
            plugin_service='kube-controllers', container_name='calico-kube-controllers')
        self.enrich_image_for_container(manifest, key,
            plugin_service='kube-controllers', container_name='calico-kube-controllers', is_init_container=False)

    def enrich_daemonset_calico_node(self, manifest: Manifest) -> None:
        """
        The method implements the enrichment procedure for Calico node DaemonSet
        :param manifest: Container to operate with manifest objects
        """

        key = "DaemonSet_calico-node"
        for container_name in ['upgrade-ipam', 'install-cni']:
            self.enrich_image_for_container(manifest, key,
                plugin_service='cni', container_name=container_name, is_init_container=True)

        self.enrich_image_for_container(manifest, key,
            plugin_service='node', container_name='mount-bpffs', is_init_container=True, allow_absent=True)
        self.enrich_image_for_container(manifest, key,
            plugin_service='flexvol', container_name='flexvol-driver', is_init_container=True, allow_absent=True)

        self.enrich_image_for_container(manifest, key,
            plugin_service='node', container_name='calico-node', is_init_container=False)

        self.enrich_daemonset_calico_node_container_env(manifest)
        self.enrich_resources_for_container(manifest, key,
            plugin_service='node', container_name='calico-node')

    def enrich_daemonset_calico_node_container_env(self, manifest: Manifest) -> None:
        """
        The method implements the enrichment procedure for 'calico-node' container in Calico node DaemonSet.
        The method attempts to preserve initial formatting.

        :param manifest: Container to operate with manifest objects
        """
        key = "DaemonSet_calico-node"
        env_delete: List[str] = []
        ip = self.inventory['services']['kubeadm']['networking']['podSubnet'].split('/')[0]
        if type(ipaddress.ip_address(ip)) is ipaddress.IPv4Address:
            env_delete.extend([
                'CALICO_IPV6POOL_CIDR', 'IP6', 'IP6_AUTODETECTION_METHOD',
                'CALICO_IPV6POOL_IPIP', 'CALICO_IPV6POOL_VXLAN'
            ])
        if not is_typha_enabled(self.inventory):
            env_delete.append('FELIX_TYPHAK8SSERVICENAME')

        env_ensure: Dict[str, str] = {
            # If metrics ports are ever configurable,
            # it makes sense to make it configurable for typha and kube-controllers as well.
            # Metrics services should also be patched.
            'FELIX_PROMETHEUSMETRICSPORT': '9091'
        }

        self.enrich_env_for_container(
            manifest, key, container_name='calico-node', env_delete=env_delete, env_ensure=env_ensure)

    def enrich_deployment_calico_typha(self, manifest: Manifest) -> None:
        """
        The method implements the enrichment procedure for Typha Deployment
        :param manifest: Container to operate with manifest objects
        """

        key = "Deployment_calico-typha"
        if not manifest.has_obj(key):
            return None
        source_yaml = manifest.get_obj(key, patch=True)

        default_tolerations = [{'key': 'node.kubernetes.io/network-unavailable', 'effect': 'NoSchedule'},
                               {'key': 'node.kubernetes.io/network-unavailable', 'effect': 'NoExecute'}]

        val = self.inventory['plugins']['calico']['typha']['replicas']
        source_yaml['spec']['replicas'] = int(val)
        self.log.verbose(f"The {key} has been patched in 'spec.replicas' with '{val}'")

        self.enrich_node_selector(manifest, key, plugin_service='typha')
        self.enrich_tolerations(manifest, key, plugin_service='typha', extra_tolerations=default_tolerations)
        self.enrich_image_for_container(manifest, key,
            plugin_service='typha', container_name='calico-typha', is_init_container=False)
        self.enrich_deployment_calico_typha_container_env(manifest)
        self.enrich_resources_for_container(manifest, key,
            plugin_service='typha', container_name='calico-typha')

    def enrich_deployment_calico_typha_container_env(self, manifest: Manifest) -> None:
        key = "Deployment_calico-typha"
        env_ensure: Dict[str, str] = {
            # If metrics ports are ever configurable, it will need to introduce new `typha.env` section.
            # Also, it makes sense to make it configurable for calico-node and kube-controllers as well.
            # Metrics services should also be patched.
            'TYPHA_PROMETHEUSMETRICSENABLED': 'true',
            'TYPHA_PROMETHEUSMETRICSPORT': '9093'
        }
        # This also searches in calico.typha.env but it is currently not supported by JSON schema.
        self.enrich_env_for_container(
            manifest, key, plugin_service='typha', container_name='calico-typha', env_ensure=env_ensure)

    def enrich_clusterrole_calico_kube_controllers(self, manifest: Manifest) -> None:
        """
        The method implements the enrichment procedure for Calico controller ClusterRole
        :param manifest: Container to operate with manifest objects
        """

        key = "ClusterRole_calico-kube-controllers"
        if self.inventory['rbac']['admission'] == "psp" and \
                self.inventory['rbac']['psp']['pod-security'] == "enabled":
            source_yaml = manifest.get_obj(key, patch=True)
            api_list = source_yaml['rules']
            api_list.append(cluster_role_use_anyuid_psp)
            self.log.verbose(f"The {key} has been patched in 'rules' with '{cluster_role_use_anyuid_psp}'")

    def enrich_clusterrole_calico_node(self, manifest: Manifest) -> None:
        """
        The method implements the enrichment procedure for Calico node ClusterRole
        :param manifest: Container to operate with manifest objects
        """

        key = "ClusterRole_calico-node"
        if self.inventory['rbac']['admission'] == "psp" and \
                self.inventory['rbac']['psp']['pod-security'] == "enabled":
            source_yaml = manifest.get_obj(key, patch=True)
            api_list = source_yaml['rules']
            api_list.append(cluster_role_use_privileged_psp)
            self.log.verbose(f"The {key} has been patched in 'rules' with '{cluster_role_use_privileged_psp}'")

    def enrich_crd_felix_configuration(self, manifest: Manifest) -> None:
        """
        The method implements the enrichment procedure for Calico CRD Felixconfigurations
        :param manifest: Container to operate with manifest objects
        """

        key = "CustomResourceDefinition_felixconfigurations.crd.projectcalico.org"
        source_yaml = manifest.get_obj(key, patch=True)

        api_list = \
        source_yaml['spec']['versions'][0]['schema']['openAPIV3Schema']['properties']['spec']['properties'][
            'prometheusMetricsEnabled']
        api_list["default"] = True
        source_yaml['spec']['versions'][0]['schema']['openAPIV3Schema']['properties']['spec']['properties'][
            'prometheusMetricsEnabled'] = api_list

    def enrich_metrics(self, manifest: Manifest) -> None:
        sz = len(manifest.all_obj_keys())
        yaml = utils.yaml_structure_preserver()
        self.include(manifest, sz, yaml.load(utils.read_internal('templates/plugins/calico-kube-controllers-metrics.yaml')))
        self.include(manifest, sz, yaml.load(utils.read_internal('templates/plugins/calico-metrics.yaml')))
        if is_typha_enabled(self.inventory):
            self.include(manifest, sz, yaml.load(utils.read_internal('templates/plugins/calico-typha-metrics.yaml')))

    def get_known_objects(self) -> List[str]:
        return [
            "ConfigMap_calico-config",
            "CustomResourceDefinition_bgpconfigurations.crd.projectcalico.org",
            "CustomResourceDefinition_bgppeers.crd.projectcalico.org",
            "CustomResourceDefinition_blockaffinities.crd.projectcalico.org",
            "CustomResourceDefinition_caliconodestatuses.crd.projectcalico.org",
            "CustomResourceDefinition_clusterinformations.crd.projectcalico.org",
            "CustomResourceDefinition_felixconfigurations.crd.projectcalico.org",
            "CustomResourceDefinition_globalnetworkpolicies.crd.projectcalico.org",
            "CustomResourceDefinition_globalnetworksets.crd.projectcalico.org",
            "CustomResourceDefinition_hostendpoints.crd.projectcalico.org",
            "CustomResourceDefinition_ipamblocks.crd.projectcalico.org",
            "CustomResourceDefinition_ipamconfigs.crd.projectcalico.org",
            "CustomResourceDefinition_ipamhandles.crd.projectcalico.org",
            "CustomResourceDefinition_ippools.crd.projectcalico.org",
            "CustomResourceDefinition_ipreservations.crd.projectcalico.org",
            "CustomResourceDefinition_kubecontrollersconfigurations.crd.projectcalico.org",
            "CustomResourceDefinition_networkpolicies.crd.projectcalico.org",
            "CustomResourceDefinition_networksets.crd.projectcalico.org",
            "ClusterRole_calico-kube-controllers",
            "ClusterRoleBinding_calico-kube-controllers",
            "ClusterRole_calico-node",
            "ClusterRoleBinding_calico-node",
            "DaemonSet_calico-node",
            "ServiceAccount_calico-node",
            "Deployment_calico-kube-controllers",
            "ServiceAccount_calico-kube-controllers",
            "PodDisruptionBudget_calico-kube-controllers",
            "Deployment_calico-typha",
            "Service_calico-typha",
            "PodDisruptionBudget_calico-typha"
        ]

    def get_enrichment_functions(self) -> List[EnrichmentFunction]:
        return [
            self.exclude_typha_objects_if_disabled,
            self.enrich_configmap_calico_config,
            self.enrich_deployment_calico_kube_controllers,
            self.enrich_daemonset_calico_node,
            self.enrich_deployment_calico_typha,
            self.enrich_clusterrole_calico_kube_controllers,
            self.enrich_clusterrole_calico_node,
            self.enrich_crd_felix_configuration,
            self.enrich_metrics,
        ]


class CalicoApiServerManifestProcessor(Processor):
    def __init__(self, logger: log.VerboseLogger, inventory: dict,
                 original_yaml_path: Optional[str] = None, destination_name: Optional[str] = None):
        super().__init__(logger, inventory, Identity('calico', 'apiserver'), original_yaml_path, destination_name)

    def get_known_objects(self) -> List[str]:
        return [
            "Namespace_calico-apiserver",
            "NetworkPolicy_allow-apiserver",
            "Service_calico-api",
            "Deployment_calico-apiserver",
            "ServiceAccount_calico-apiserver",
            "APIService_v3.projectcalico.org",
            "ClusterRole_calico-crds",
            "ClusterRole_calico-extension-apiserver-auth-access",
            "ClusterRole_calico-webhook-reader",
            "ClusterRoleBinding_calico-apiserver-access-crds",
            "ClusterRoleBinding_calico-apiserver-delegate-auth",
            "ClusterRoleBinding_calico-apiserver-webhook-reader",
            "ClusterRoleBinding_calico-extension-apiserver-auth-access",
        ]

    def get_enrichment_functions(self) -> List[EnrichmentFunction]:
        return [
            self.enrich_namespace_calico_apiserver,
            self.enrich_deployment_calico_apiserver,
            self.enrich_clusterrole_calico_crds,
        ]

    def get_namespace_to_necessary_pss_profiles(self) -> Dict[str, str]:
        return {'calico-apiserver': 'baseline'}

    def enrich_namespace_calico_apiserver(self, manifest: Manifest) -> None:
        self.assign_default_pss_labels(manifest, 'calico-apiserver')

    def enrich_deployment_calico_apiserver(self, manifest: Manifest) -> None:
        key = "Deployment_calico-apiserver"
        self.enrich_node_selector(manifest, key, plugin_service='apiserver')
        self.enrich_tolerations(manifest, key, plugin_service='apiserver')
        self.enrich_image_for_container(
            manifest, key, plugin_service='apiserver', container_name='calico-apiserver', is_init_container=False)
        self.enrich_resources_for_container(
            manifest, key, plugin_service='apiserver', container_name='calico-apiserver')

        self.enrich_deployment_calico_apiserver_container(manifest)

    def enrich_deployment_calico_apiserver_container(self, manifest: Manifest) -> None:
        key = "Deployment_calico-apiserver"
        # By default, API server searches in the same directory as specified below,
        # but with different file names: apiserver.crt, apiserver.key
        # There is no real necessity to change paths
        # except to make TLS secret creation process the same as for nginx-ingress.
        additional_args = [
            "--tls-cert-file=apiserver.local.config/certificates/tls.crt",
            "--tls-private-key-file=apiserver.local.config/certificates/tls.key"
        ]
        # This also searches in calico.apiserver.args but it is currently not supported by JSON schema.
        self.enrich_args_for_container(
            manifest, key, plugin_service='apiserver', container_name='calico-apiserver',
            extra_args=additional_args)

    def enrich_clusterrole_calico_crds(self, manifest: Manifest) -> None:
        key = "ClusterRole_calico-crds"
        if self.inventory['rbac']['admission'] == "psp" and \
                self.inventory['rbac']['psp']['pod-security'] == "enabled":
            source_yaml = manifest.get_obj(key, patch=True)
            api_list = source_yaml['rules']
            api_list.append(cluster_role_use_anyuid_psp)
            self.log.verbose(f"The {key} has been patched in 'rules' with '{cluster_role_use_anyuid_psp}'")


cluster_role_use_anyuid_psp = {
        "apiGroups": ["policy"],
        "resources": ["podsecuritypolicies"],
        "verbs":     ["use"],
        "resourceNames": ["oob-anyuid-psp"]
}

cluster_role_use_privileged_psp = {
        "apiGroups": ["policy"],
        "resources": ["podsecuritypolicies"],
        "verbs":     ["use"],
        "resourceNames": ["oob-privileged-psp"]
}



#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import OrderedDict
from itertools import chain
from typing import List

from kubemarine import kubernetes, plugins, admission
from kubemarine.core import flow
from kubemarine.core import utils
from kubemarine.core.action import Action
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.resources import DynamicResources
from kubemarine.procedures import install


def system_prepare_thirdparties(cluster: KubernetesCluster) -> None:
    if not cluster.inventory['services'].get('thirdparties', {}):
        cluster.log.debug("Skipped - no thirdparties defined in config file")
        return

    install.system_prepare_thirdparties(cluster)


def prepull_images(cluster: KubernetesCluster) -> None:
    cluster.log.debug("Prepulling Kubernetes images...")
    fix_cri_socket(cluster)
    upgrade_group = kubernetes.get_group_for_upgrade(cluster)
    upgrade_group.call(kubernetes.images_grouped_prepull)


def kubernetes_upgrade(cluster: KubernetesCluster) -> None:
    minor_version = int(cluster.context['upgrade_version'].split('.')[1])

    upgrade_group = kubernetes.get_group_for_upgrade(cluster)
    if minor_version >= 28 and cluster.inventory['rbac']['pss']['pod-security'] == 'enabled':
        first_control_plane = cluster.nodes["control-plane"].get_first_member()

        cluster.log.debug("Updating kubeadm config map")
        final_features_list = first_control_plane.call(admission.update_kubeadm_configmap_pss, target_state="enabled")
        cluster.log.debug("Updating kube-apiserver configs on control-planes")
        cluster.nodes["control-plane"].call(admission.update_kubeapi_config_pss, features_list=final_features_list)

    drain_timeout = cluster.procedure_inventory.get('drain_timeout')
    grace_period = cluster.procedure_inventory.get('grace_period')
    disable_eviction = cluster.procedure_inventory.get("disable-eviction", True)
    drain_kwargs = {
        'disable_eviction': disable_eviction, 'drain_timeout': drain_timeout, 'grace_period': grace_period
    }

    kubernetes.upgrade_first_control_plane(upgrade_group, cluster, **drain_kwargs)

    # After first control-plane upgrade is finished we may loose our CoreDNS changes.
    # Thus, we need to re-apply our CoreDNS changes immediately after first control-plane upgrade.
    install.deploy_coredns(cluster)

    kubernetes.upgrade_other_control_planes(upgrade_group, cluster, **drain_kwargs)

    if cluster.nodes.get('worker', []):
        kubernetes.upgrade_workers(upgrade_group, cluster, **drain_kwargs)

    cluster.nodes['control-plane'].get_first_member().sudo('rm -f /etc/kubernetes/nodes-k8s-versions.txt')
    cluster.context['cached_nodes_versions_cleaned'] = True


def kubernetes_cleanup_nodes_versions(cluster: KubernetesCluster) -> None:
    if not cluster.context.get('cached_nodes_versions_cleaned', False):
        cluster.log.verbose('Cached nodes versions required')
        cluster.nodes['control-plane'].get_first_member().sudo('rm -f /etc/kubernetes/nodes-k8s-versions.txt')
    else:
        cluster.log.verbose('Cached nodes versions already cleaned')
    kubernetes_apply_taints(cluster)


def upgrade_packages(cluster: KubernetesCluster) -> None:
    upgrade_version = cluster.context["upgrade_version"]

    packages = cluster.procedure_inventory.get(upgrade_version, {}).get("packages", {})
    if packages.get("install") is not None or packages.get("upgrade") is not None or packages.get("remove") is not None:
        install.manage_custom_packages(cluster.nodes['all'])


def upgrade_plugins(cluster: KubernetesCluster) -> None:
    upgrade_version = cluster.context["upgrade_version"]

    # upgrade_candidates is a source of upgradeable plugins, not list of plugins to upgrade.
    # Some plugins from upgrade_candidates will not be upgraded, because they have "install: false"
    upgrade_candidates = {}
    defined_plugins = cluster.procedure_inventory.get(upgrade_version, {}).get("plugins", {}).keys()
    for plugin in chain(defined_plugins, plugins.oob_plugins):
        # TODO: use only OOB plugins that have changed version so that we do not perform redundant installations
        upgrade_candidates[plugin] = cluster.inventory["plugins"][plugin]

    plugins.install(cluster, upgrade_candidates)


tasks = OrderedDict({
    "verify_upgrade_versions": kubernetes.verify_upgrade_versions,
    "thirdparties": system_prepare_thirdparties,
    "prepull_images": prepull_images,
    "kubernetes": kubernetes_upgrade,
    "kubernetes_cleanup": kubernetes_cleanup_nodes_versions,
    "packages": upgrade_packages,
    "plugins": upgrade_plugins,
    "overview": install.overview

})


class UpgradeFlow(flow.Flow):
    def __init__(self) -> None:
        self.target_version = "not supported"

    def _run(self, resources: DynamicResources) -> None:
        logger = resources.logger()

        previous_version = kubernetes.get_initial_kubernetes_version(resources.raw_inventory())
        upgrade_plan = resources.procedure_inventory().get('upgrade_plan')
        if not upgrade_plan:
            raise Exception('Upgrade plan is not specified in procedure')
        upgrade_plan = verify_upgrade_plan(previous_version, upgrade_plan)
        logger.debug(f"Loaded upgrade plan: current ({previous_version}) ⭢ {' ⭢ '.join(upgrade_plan)}")

        self.target_version = upgrade_plan[-1]
        kubernetes.verify_supported_version(self.target_version, logger)

        args = resources.context['execution_arguments']
        if (args['tasks'] or args['exclude']) and len(upgrade_plan) > 1:
            raise Exception("Usage of '--tasks' and '--exclude' is not allowed when upgrading to more than one version")

        # todo inventory is preserved few times, probably need to preserve it once instead.
        actions = [UpgradeAction(version) for version in upgrade_plan]
        flow.run_actions(resources, actions)


class UpgradeAction(Action):
    def __init__(self, upgrade_version: str) -> None:
        super().__init__('upgrade to ' + upgrade_version, recreate_inventory=True)
        self.upgrade_version = upgrade_version

    def run(self, res: DynamicResources) -> None:
        flow.run_tasks(res, tasks)
        res.make_final_inventory()

    def prepare_context(self, context: dict) -> None:
        context['upgrade_version'] = self.upgrade_version
        context['dump_filename_prefix'] = self.upgrade_version


def create_context(cli_arguments: List[str] = None) -> dict:
    cli_help = '''
    Script for automated upgrade of the entire Kubernetes cluster to a new version.

    How to use:

    '''

    parser = flow.new_procedure_parser(cli_help, tasks=tasks)

    context = flow.create_context(parser, cli_arguments, procedure='upgrade')
    return context


def main(cli_arguments: List[str] = None) -> None:
    context = create_context(cli_arguments)
    flow_ = UpgradeFlow()
    result = flow_.run_flow(context)

    kubernetes.verify_supported_version(flow_.target_version, result.logger)


def verify_upgrade_plan(previous_version: str, upgrade_plan: List[str]) -> List[str]:
    kubernetes.verify_allowed_version(previous_version)
    for version in upgrade_plan:
        kubernetes.verify_allowed_version(version)

    upgrade_plan.sort(key=utils.version_key)

    for version in upgrade_plan:
        kubernetes.test_version_upgrade_possible(previous_version, version)
        previous_version = version

    return upgrade_plan


def fix_cri_socket(cluster: KubernetesCluster) -> None:
    """
    This method fixs the issue with 'kubeadm.alpha.kubernetes.io/cri-socket' node annotation
    and delete the docker socket if it exists
    """

    if cluster.inventory["services"]["cri"]["containerRuntime"] == "containerd":
        control_plane = cluster.nodes["control-plane"].get_first_member()
        control_plane.sudo(f"sudo kubectl annotate nodes --all "
                           f"--overwrite kubeadm.alpha.kubernetes.io/cri-socket=/run/containerd/containerd.sock")
        upgrade_group = kubernetes.get_group_for_upgrade(cluster)
        upgrade_group.sudo("rm -rf /var/run/docker.sock")


def kubernetes_apply_taints(cluster: KubernetesCluster) -> None:
    # Apply taints after upgrade
    group = cluster.nodes['control-plane']
    kubernetes.apply_taints(group)


if __name__ == '__main__':
    main()



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import ipaddress
from typing import Optional, List, Dict

from kubemarine.core import utils, log
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.group import NodeGroup
from kubemarine.kubernetes import secrets
from kubemarine.plugins.manifest import Processor, EnrichmentFunction, Manifest, Identity


def check_job_for_nginx(cluster: KubernetesCluster) -> None:
    first_control_plane = cluster.nodes['control-plane'].get_first_member()
    version = cluster.inventory['plugins']['nginx-ingress-controller']['version'].replace('v', '.').split('.')

    major_version = int(version[1])
    minor_version = int(version[2])

    check_jobs = first_control_plane.sudo(f"kubectl get jobs -n ingress-nginx")
    if list(check_jobs.values())[0].stderr == "" and major_version >= 1 and minor_version >= 4:
        cluster.log.debug('Delete old jobs for nginx')
        first_control_plane.sudo(f"sudo kubectl delete job --all -n ingress-nginx")
    else:
        cluster.log.debug('There are no jobs to delete')


def enrich_inventory(inventory: dict, cluster: KubernetesCluster) -> dict:
    if not inventory["plugins"]["nginx-ingress-controller"]["install"]:
        return inventory

    # Change type for hostPorts because of jinja enrichment
    for port in inventory["plugins"]["nginx-ingress-controller"].get("ports", []):
        if "hostPort" in port and not isinstance(port['hostPort'], int):
            port['hostPort'] = int(port['hostPort'])

    if inventory["plugins"]["nginx-ingress-controller"].get('custom_headers'):
        if not inventory["plugins"]["nginx-ingress-controller"].get('config_map'):
            inventory["plugins"]["nginx-ingress-controller"]['config_map'] = {}
        if not inventory["plugins"]["nginx-ingress-controller"]['config_map'].get('proxy-set-headers'):
            inventory["plugins"]["nginx-ingress-controller"]['config_map']['proxy-set-headers'] = 'ingress-nginx/custom-headers'

    # if user defined resources himself, we should use them as is, instead of merging with our defaults
    raw_controller = cluster.raw_inventory.get("plugins", {}).get("nginx-ingress-controller", {}).get("controller", {})
    if "resources" in raw_controller:
        inventory["plugins"]["nginx-ingress-controller"]["controller"]["resources"] = raw_controller["resources"]
    raw_webhook = cluster.raw_inventory.get("plugins", {}).get("nginx-ingress-controller", {}).get("webhook", {})
    if "resources" in raw_webhook:
        inventory["plugins"]["nginx-ingress-controller"]["webhook"]["resources"] = raw_webhook["resources"]

    return inventory


def cert_renew_enrichment(inventory: dict, cluster: KubernetesCluster) -> dict:
    # check that renewal is required for nginx
    if cluster.context.get('initial_procedure') != 'cert_renew' \
            or not cluster.procedure_inventory.get("nginx-ingress-controller"):
        return inventory

    nginx_plugin = inventory["plugins"]["nginx-ingress-controller"]

    # check that renewal is possible
    if not nginx_plugin["install"]:
        raise Exception("Certificates can not be renewed for nginx plugin since it is not installed")

    # update certificates in inventory
    nginx_plugin["controller"]["ssl"]["default-certificate"] = cluster.procedure_inventory["nginx-ingress-controller"]

    return inventory


def finalize_inventory(cluster: KubernetesCluster, inventory_to_finalize: dict) -> dict:
    # check that renewal is required for nginx
    if cluster.context.get('initial_procedure') != 'cert_renew' \
            or not cluster.procedure_inventory.get("nginx-ingress-controller"):
        return inventory_to_finalize

    if not inventory_to_finalize["plugins"].get("nginx-ingress-controller"):
        inventory_to_finalize["plugins"]["nginx-ingress-controller"] = {}

    if not inventory_to_finalize["plugins"]["nginx-ingress-controller"].get("controller"):
        inventory_to_finalize["plugins"]["nginx-ingress-controller"]["controller"] = {}

    if not inventory_to_finalize["plugins"]["nginx-ingress-controller"]["controller"].get("ssl"):
        inventory_to_finalize["plugins"]["nginx-ingress-controller"]["controller"]["ssl"] = {}

    nginx_plugin = inventory_to_finalize["plugins"]["nginx-ingress-controller"]
    nginx_plugin["controller"]["ssl"]["default-certificate"] = cluster.procedure_inventory["nginx-ingress-controller"]

    return inventory_to_finalize


def redeploy_ingress_nginx_is_needed(cluster: KubernetesCluster) -> bool:
    # redeploy ingres-nginx-controller for add/remove node procedures is needed in case:
    # 1. plugins.nginx-ingress-controller.install=true
    # 2. any balancer node exists (including as remove_node)
    # 3. all balancers have add_node/remove_node roles (added the first or removed the last balancer)
    # 4. One of following is not overriden:
    #    4.1. use-proxy-protocol
    #    4.2. ingress-nginx-ports and some from target ports
    ingress_nginx_plugin = cluster.inventory['plugins']['nginx-ingress-controller']
    balancers = [balancer for balancer in cluster.inventory['nodes'] if 'balancer' in balancer['roles']]
    if not ingress_nginx_plugin.get("install", False) or \
            not balancers or \
            any('add_node' not in node['roles'] and 'remove_node' not in node['roles'] for node in balancers):
        return False

    proxy_protocol_overriden = 'use-proxy-protocol' in cluster.raw_inventory.get('plugins', {})\
        .get('nginx-ingress-controller', {})\
        .get('config_map', {})
    http_target_port_overriden = 'http' in cluster.raw_inventory.get('services', {})\
        .get('loadbalancer', {})\
        .get('target_ports', {})
    https_target_port_overriden = 'https' in cluster.raw_inventory.get('services', {}) \
        .get('loadbalancer', {}) \
        .get('target_ports', {})
    ingress_nginx_ports_overriden = 'ports' in cluster.raw_inventory.get('plugins', {})\
        .get('nginx-ingress-controller', {})
    return not proxy_protocol_overriden or not (ingress_nginx_ports_overriden or
                                                (https_target_port_overriden and http_target_port_overriden))


def manage_custom_certificate(cluster: KubernetesCluster) -> None:
    default_cert = cluster.inventory["plugins"]["nginx-ingress-controller"]["controller"]["ssl"]\
        .get("default-certificate")
    if not default_cert:
        cluster.log.debug("No custom default ingress certificate specified, skipping...")
        return

    secret_name = "default-ingress-cert"
    secret_namespace = "kube-system"

    first_control_plane = cluster.nodes["control-plane"].get_first_member()

    # first, we need to load cert and key files to first control-plane to known locations
    with secrets.create_tls_secret_procedure(first_control_plane):
        first_control_plane.call(put_custom_certificate, default_cert=default_cert)

        # second, we need to validate cert and key using openssl
        first_control_plane.call(verify_certificate_and_key)

        # third, we need to create tls secret under well-known name
        # this certificate is already configured to be used by controller
        first_control_plane.call(secrets.renew_tls_secret,
                                 name=secret_name,
                                 namespace=secret_namespace)
    # fourth, we need to remove base path dir when the procedure is exited


def put_custom_certificate(first_control_plane: NodeGroup, default_cert: dict) -> None:
    if default_cert.get("data"):
        cert = io.StringIO(default_cert["data"]["cert"])
        key = io.StringIO(default_cert["data"]["key"])
    else:
        cert = io.StringIO(utils.read_external(default_cert["paths"]["cert"]))
        key = io.StringIO(utils.read_external(default_cert["paths"]["key"]))

    secrets.put_certificate(first_control_plane, cert, key)


def verify_certificate_and_key(first_control_plane: NodeGroup) -> None:
    if not secrets.verify_certificate(first_control_plane):
        raise Exception("Custom default ingress certificate and key are not compatible!")


class IngressNginxManifestProcessor(Processor):
    def __init__(self, logger: log.VerboseLogger, inventory: dict,
                 original_yaml_path: Optional[str] = None, destination_name: Optional[str] = None) -> None:
        super().__init__(logger, inventory, Identity('nginx-ingress-controller'), original_yaml_path, destination_name)

    def get_known_objects(self) -> List[str]:
        return [
            "Namespace_ingress-nginx",
            "ServiceAccount_ingress-nginx",
            "ServiceAccount_ingress-nginx-admission",
            "Role_ingress-nginx",
            "Role_ingress-nginx-admission",
            "ClusterRole_ingress-nginx",
            "ClusterRole_ingress-nginx-admission",
            "RoleBinding_ingress-nginx",
            "RoleBinding_ingress-nginx-admission",
            "ClusterRoleBinding_ingress-nginx",
            "ClusterRoleBinding_ingress-nginx-admission",
            "ConfigMap_ingress-nginx-controller",
            "Service_ingress-nginx-controller",
            "Service_ingress-nginx-controller-admission",
            "Deployment_ingress-nginx-controller",
            "Job_ingress-nginx-admission-create",
            "Job_ingress-nginx-admission-patch",
            "IngressClass_nginx",
            "ValidatingWebhookConfiguration_ingress-nginx-admission",
        ]

    def get_enrichment_functions(self) -> List[EnrichmentFunction]:
        return [
            self.enrich_namespace_ingress_nginx,
            self.enrich_configmap_ingress_nginx_controller,
            self.add_configmap_ingress_nginx_controller,
            self.enrich_deployment_ingress_nginx_controller,
            self.enrich_ingressclass_nginx,
            self.enrich_job_ingress_nginx_admission_create,
            self.enrich_job_ingress_nginx_admission_patch,
            self.enrich_service_ingress_nginx_controller,
        ]

    def get_namespace_to_necessary_pss_profiles(self) -> Dict[str, str]:
        return {'ingress-nginx': 'privileged'}

    def enrich_namespace_ingress_nginx(self, manifest: Manifest) -> None:
        self.assign_default_pss_labels(manifest, 'ingress-nginx')

    def enrich_configmap_ingress_nginx_controller(self, manifest: Manifest) -> None:
        key = "ConfigMap_ingress-nginx-controller"
        config_map = self.inventory['plugins']['nginx-ingress-controller'].get('config_map')
        if config_map:
            source_yaml = manifest.get_obj(key, patch=True)
            data: dict = source_yaml['data']
            data.update(config_map)
            self.log.verbose(f"The {key} has been patched in 'data' "
                             f"with the data from 'plugins.nginx-ingress-controller.config_map'")

    def add_configmap_ingress_nginx_controller(self, manifest: Manifest) -> None:
        custom_headers = self.inventory['plugins']['nginx-ingress-controller'].get('custom_headers')
        if custom_headers:
            custom_headers_cm = dict(CUSTOM_HEADERS_CM)
            custom_headers_cm['data'] = custom_headers
            # Insert custom-headers ConfigMap before ingress-nginx-controller ConfigMap
            ingres_nginx_cm = manifest.key_index("ConfigMap_ingress-nginx-controller")
            self.include(manifest, ingres_nginx_cm, custom_headers_cm)
            self.log.verbose(f"The {manifest.obj_key(custom_headers_cm)} has been patched in 'data' "
                             f"with the data from 'plugins.nginx-ingress-controller.custom_headers'")

    def enrich_deployment_ingress_nginx_controller(self, manifest: Manifest) -> None:
        key = "Deployment_ingress-nginx-controller"
        source_yaml = manifest.get_obj(key, patch=True)

        self.enrich_deamonset_ingress_nginx_controller_container(manifest)

        self.enrich_image_for_container(manifest, key,
            plugin_service='controller', container_name='controller', is_init_container=False)

        self.enrich_resources_for_container(manifest, key, container_name='controller', plugin_service="controller")
        self.enrich_node_selector(manifest, key, plugin_service='controller')
        self.enrich_tolerations(manifest, key, plugin_service='controller')

        # Patch kind in the last step to avoid sudden key change in log messages
        source_yaml['kind'] = 'DaemonSet'
        self.log.verbose(f"The {key} has been patched in 'kind' with 'DaemonSet'")

    def enrich_deamonset_ingress_nginx_controller_container(self, manifest: Manifest) -> None:
        key = "Deployment_ingress-nginx-controller"

        extra_args: List[str] = [
            '--watch-ingress-without-class=true'
        ]
        ssl_options = self.inventory['plugins']['nginx-ingress-controller']['controller']['ssl']

        if ssl_options['enableSslPassthrough']:
            extra_args.append('--enable-ssl-passthrough')
        if ssl_options.get('default-certificate'):
            extra_args.append('--default-ssl-certificate' + '=' + 'kube-system/default-ingress-cert')

        self.enrich_deamonset_ingress_nginx_controller_container_args(
            manifest, remove_args=['--publish-service'], extra_args=extra_args)

        container_pos, container = self.find_container_for_patch(
            manifest, key, container_name='controller', is_init_container=False)

        container['ports'] = self.inventory['plugins']['nginx-ingress-controller']['ports']
        self.log.verbose(f"The {key} has been patched in 'spec.template.spec.containers.[{container_pos}].ports' "
                         f"with the data from 'plugins.nginx-ingress-controller.ports'")

    def enrich_deamonset_ingress_nginx_controller_container_args(self, manifest: Manifest,
                                                                 *,
                                                                 remove_args: List[str],
                                                                 extra_args: List[str]) -> None:
        key = "Deployment_ingress-nginx-controller"
        self.enrich_args_for_container(manifest, key,
                                       plugin_service='controller', container_name='controller',
                                       remove_args=remove_args,
                                       extra_args=extra_args)

    def enrich_ingressclass_nginx(self, manifest: Manifest) -> None:
        key = "IngressClass_nginx"
        source_yaml = manifest.get_obj(key, patch=True)
        source_yaml['metadata'].setdefault('annotations', {})['ingressclass.kubernetes.io/is-default-class'] = 'true'
        self.log.verbose(f"The {key} has been patched in 'metadata.annotations' "
                         f"with 'ingressclass.kubernetes.io/is-default-class: true'")

    def enrich_job_ingress_nginx_admission_create(self, manifest: Manifest) -> None:
        key = "Job_ingress-nginx-admission-create"
        self.enrich_image_for_container(manifest, key,
            plugin_service='webhook', container_name='create', is_init_container=False)

        self.enrich_resources_for_container(manifest, key, container_name='create', plugin_service="webhook")

    def enrich_job_ingress_nginx_admission_patch(self, manifest: Manifest) -> None:
        key = "Job_ingress-nginx-admission-patch"
        self.enrich_image_for_container(manifest, key,
            plugin_service='webhook', container_name='patch', is_init_container=False)

        self.enrich_resources_for_container(manifest, key, container_name='patch', plugin_service="webhook")

    def enrich_service_ingress_nginx_controller(self, manifest: Manifest) -> None:
        # The method needs some rework in case of dual stack support
        key = "Service_ingress-nginx-controller"
        ip = self.inventory['services']['kubeadm']['networking']['serviceSubnet'].split('/')[0]
        if type(ipaddress.ip_address(ip)) is ipaddress.IPv6Address:
            source_yaml = manifest.get_obj(key, patch=True)
            source_yaml['spec']['ipFamilies'] = ['IPv6']
            self.log.verbose(f"The {key} has been patched in 'spec.ipFamilies' with 'IPv6'")


class V1_2_X_IngressNginxManifestProcessor(IngressNginxManifestProcessor):
    def get_enrichment_functions(self) -> List[EnrichmentFunction]:
        enrichment_functions = super().get_enrichment_functions()
        enrichment_functions.extend([
            self.exclude_webhook_resources,
            self.enrich_role_ingress_nginx,
        ])
        return enrichment_functions

    def enrich_deamonset_ingress_nginx_controller_container(self, manifest: Manifest) -> None:
        key = "Deployment_ingress-nginx-controller"

        container_pos, container = self.find_container_for_patch(
            manifest, key, container_name='controller', is_init_container=False)

        del container['volumeMounts']
        self.log.verbose(f"The 'volumeMounts' property has been removed "
                         f"from 'spec.template.spec.containers.[{container_pos}]' in the {key}")

        super().enrich_deamonset_ingress_nginx_controller_container(manifest)

    def enrich_deamonset_ingress_nginx_controller_container_args(self, manifest: Manifest,
                                                                 *,
                                                                 remove_args: List[str],
                                                                 extra_args: List[str]) -> None:
        webhook_args_remove = [
            '--validating-webhook',
            '--validating-webhook-certificate',
            '--validating-webhook-key'
        ]
        remove_args = remove_args + webhook_args_remove
        super().enrich_deamonset_ingress_nginx_controller_container_args(
            manifest, remove_args=remove_args, extra_args=extra_args)

    def enrich_job_ingress_nginx_admission_create(self, manifest: Manifest) -> None:
        return

    def enrich_job_ingress_nginx_admission_patch(self, manifest: Manifest) -> None:
        return

    def exclude_webhook_resources(self, manifest: Manifest) -> None:
        webhook_resources = [
            "ServiceAccount_ingress-nginx-admission",
            "Role_ingress-nginx-admission",
            "ClusterRole_ingress-nginx-admission",
            "RoleBinding_ingress-nginx-admission",
            "ClusterRoleBinding_ingress-nginx-admission",
            "Service_ingress-nginx-controller-admission",
            "Job_ingress-nginx-admission-create",
            "Job_ingress-nginx-admission-patch",
            "ValidatingWebhookConfiguration_ingress-nginx-admission",
        ]
        for key in webhook_resources:
            self.exclude(manifest, key)

    def enrich_role_ingress_nginx(self, manifest: Manifest) -> None:
        key = "Role_ingress-nginx"
        source_yaml = manifest.get_obj(key, patch=True)
        # TODO patch only if psp is enabled?
        api_list = source_yaml['rules']
        api_list.append(psp_ingress_nginx)
        self.log.verbose(f"The {key} has been patched in 'rules' with {psp_ingress_nginx}")


def get_ingress_nginx_manifest_processor(logger: log.VerboseLogger, inventory: dict,
                                         yaml_path: Optional[str] = None, destination: Optional[str] = None) -> Processor:
    version: str = inventory['plugins']['nginx-ingress-controller']['version']
    kwargs = {'original_yaml_path': yaml_path, 'destination_name': destination}
    if utils.minor_version(version) == 'v1.2':
        return V1_2_X_IngressNginxManifestProcessor(logger, inventory, **kwargs)

    return IngressNginxManifestProcessor(logger, inventory, **kwargs)


CUSTOM_HEADERS_CM = {
    "apiVersion": "v1",
    "kind": "ConfigMap",
    "metadata": {
        "name": "custom-headers",
        "namespace": "ingress-nginx"
    }
}

psp_ingress_nginx = {
    "apiGroups": ["extensions"],
    "resources": ["podsecuritypolicies"],
    "verbs":     ["use"],
    "resourceNames": ["oob-host-network-psp"]
}



#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import OrderedDict
from typing import List

from kubemarine.core import flow
from kubemarine.core.action import Action
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.resources import DynamicResources
from kubemarine.procedures import install
from kubemarine import system


def reboot(cluster: KubernetesCluster) -> None:
    if cluster.context.get('initial_procedure') != 'reboot':
        raise ImportError('Invalid reboot.py usage, please use system.reboot_nodes')

    if not cluster.procedure_inventory.get("nodes"):
        cluster.log.verbose('No nodes defined in procedure: all nodes will be rebooted')
    else:
        cluster.log.verbose('There are nodes defined in procedure: only defined will be rebooted')

    nodes = []

    cluster.log.verbose('The following nodes will be rebooted:')
    for node in cluster.procedure_inventory.get("nodes", cluster.inventory['nodes']):
        nodes.append(node['name'])
        cluster.log.verbose('  - ' + node['name'])

    system.reboot_group(cluster.make_group_from_nodes(nodes),
                        try_graceful=cluster.procedure_inventory.get("graceful_reboot"))


tasks = OrderedDict({
    "reboot": reboot,
    "overview": install.overview,
})


class RebootAction(Action):
    def __init__(self) -> None:
        super().__init__('reboot')

    def run(self, res: DynamicResources) -> None:
        flow.run_tasks(res, tasks)


def create_context(cli_arguments: List[str] = None) -> dict:
    cli_help = '''
    Script for Kubernetes nodes graceful rebooting.

    How to use:

    '''

    parser = flow.new_procedure_parser(cli_help, optional_config=True, tasks=tasks)

    context = flow.create_context(parser, cli_arguments, procedure='reboot')
    return context


def main(cli_arguments: List[str] = None) -> None:
    context = create_context(cli_arguments)
    flow.ActionsFlow([RebootAction()]).run_flow(context)


if __name__ == '__main__':
    main()



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import deepcopy
from typing import Dict, List, Union, Iterable, Tuple, Optional, Any, Callable

import yaml

from kubemarine.core import log, utils, static, connections
from kubemarine.core.connections import ConnectionPool
from kubemarine.core.environment import Environment
from kubemarine.core.group import NodeGroup, NodeConfig

_AnyConnectionTypes = Union[str, NodeGroup]


class KubernetesCluster(Environment):

    def __init__(self, inventory: dict, context: dict, procedure_inventory: dict = None,
                 logger: log.EnhancedLogger = None) -> None:
        self.roles: List[str] = []
        self.ips: Dict[str, List[str]] = {
            "all": []
        }
        self.nodes: Dict[str, NodeGroup] = {}

        self.raw_inventory = deepcopy(inventory)
        # Should not be copied. Can be used after successful of failed execution and might store intermediate result.
        self.context = context
        self.procedure_inventory = {} if procedure_inventory is None else deepcopy(procedure_inventory)

        self._connection_pool: ConnectionPool = connections.EMPTY_POOL

        self._logger = logger if logger is not None \
            else log.init_log_from_context_args(self.globals, self.context, self.raw_inventory).logger

        self._inventory: dict = {}

    def enrich(self, custom_enrichment_fns: List[str] = None) -> None:
        # do not make dumps for custom enrichment functions, because result is generally undefined
        make_dumps = custom_enrichment_fns is None
        from kubemarine.core import defaults
        self._inventory = defaults.enrich_inventory(
            self, deepcopy(self.raw_inventory), make_dumps=make_dumps, enrichment_functions=custom_enrichment_fns)

        self._connection_pool = self.create_connection_pool(self.ips['all'])

    @property
    def inventory(self) -> dict:
        return self._inventory

    @property
    def connection_pool(self) -> ConnectionPool:
        return self._connection_pool

    def create_connection_pool(self, hosts: List[str]) -> ConnectionPool:
        return ConnectionPool(self.inventory, hosts)

    @property
    def log(self) -> log.EnhancedLogger:
        return self._logger

    @property
    def globals(self) -> dict:
        return static.GLOBALS

    def make_group(self, ips: Iterable[_AnyConnectionTypes]) -> NodeGroup:
        return NodeGroup(ips, self)

    def get_access_address_from_node(self, node: NodeConfig) -> str:
        """
        Returns address which should be used to connect to the node via Fabric.
        The address also can be used as unique identifier of the node.
        """
        address: str
        if node.get('connect_to') is not None:
            address = node['connect_to']
        elif node.get('address') is not None:
            address = node['address']
        else:
            address = node['internal_address']

        return address

    def get_nodes_by_names(self, node_names: List[str]) -> List[dict]:
        result = []
        for node in self.inventory["nodes"]:
            if node['name'] in node_names:
                result.append(node)

        return result

    def get_node_by_name(self, node_name: str) -> Optional[dict]:
        nodes = self.get_nodes_by_names([node_name])
        return next(iter(nodes), None)

    def get_addresses_from_node_names(self, node_names: List[str]) -> List[str]:
        return [self.get_access_address_from_node(node)
                for node in self.get_nodes_by_names(node_names)]

    def get_node(self, host: _AnyConnectionTypes) -> NodeConfig:
        return self.make_group([host]).get_config()

    def make_group_from_nodes(self, node_names: List[str]) -> NodeGroup:
        ips = self.get_addresses_from_node_names(node_names)
        return self.make_group(ips)

    def make_group_from_roles(self, roles: Iterable[str]) -> NodeGroup:
        group = self.make_group([])
        for role in roles:
            if role not in self.nodes:
                self.log.verbose(f'Group {role!r} is requested for usage, but this group does not exist.')
                continue

            group = group.include_group(self.nodes[role])

        return group

    def create_group_from_groups_nodes_names(self, groups_names: List[str], nodes_names: List[str]) -> NodeGroup:
        common_group = self.make_group_from_roles(groups_names)

        if nodes_names:
            common_group = common_group.include_group(self.make_group_from_nodes(nodes_names))

        return common_group

    def schedule_cumulative_point(self, point_method: Callable) -> None:
        from kubemarine.core import flow
        flow.schedule_cumulative_point(self, point_method)

    def is_task_completed(self, task_path: str) -> bool:
        from kubemarine.core import flow
        return flow.is_task_completed(self, task_path)

    def get_facts_enrichment_fns(self) -> List[str]:
        return [
            "kubemarine.core.schema.verify_inventory",
            "kubemarine.core.defaults.merge_defaults",
            "kubemarine.kubernetes.verify_initial_version",
            "kubemarine.kubernetes.add_node_enrichment",
            "kubemarine.core.defaults.calculate_node_names",
            "kubemarine.kubernetes.remove_node_enrichment",
            "kubemarine.controlplane.controlplane_node_enrichment",
            "kubemarine.core.defaults.append_controlplain",
            "kubemarine.core.defaults.compile_inventory",
            "kubemarine.core.defaults.verify_node_names",
            "kubemarine.core.defaults.apply_defaults",
            "kubemarine.core.defaults.calculate_nodegroups"
        ]

    def detect_nodes_context(self) -> dict:
        """The method should fetch only node specific information that is not changed during Kubemarine run"""
        self.log.debug('Start detecting nodes context...')

        from kubemarine import system
        system.whoami(self)
        self.log.verbose('Whoami check finished')

        self._check_online_nodes()
        self._check_accessible_nodes()

        system.detect_active_interface(self)
        self.log.verbose('Interface check finished')
        system.detect_os_family(self)
        self.log.verbose('OS family check finished')

        self.log.debug('Detecting nodes context finished!')
        return deepcopy(self.context['nodes'])

    def _check_online_nodes(self) -> None:
        """
        Check that only subset of nodes for removal can be offline
        """
        all = self.nodes['all']
        for_removal = all.get_nodes_for_removal()
        remained = all.exclude_group(for_removal)
        offline = all.get_online_nodes(False)
        remained_offline = remained.intersection_group(offline)
        if not remained_offline.is_empty():
            raise Exception(f"{remained_offline.get_hosts()} are not reachable. "
                            "Probably they are turned off or something is incorrect with ssh daemon, "
                            "or incorrect ssh port is specified.")

    # todo this check can probably be moved to prepare.check tasks group of each procedure
    def _check_accessible_nodes(self) -> None:
        """
        Check that all online nodes are accessible.
        """
        all = self.nodes['all']
        online = all.get_online_nodes(True)
        accessible = all.get_accessible_nodes()
        not_accessible = all.exclude_group(accessible)
        not_accessible_online = online.intersection_group(not_accessible)
        if not not_accessible_online.is_empty():
            raise Exception(f"{not_accessible_online.get_hosts()} are not accessible through ssh. "
                            f"Check ssh credentials.")

    def get_os_family_for_node(self, host: str) -> str:
        os_family: Optional[str] = self.context['nodes'].get(host, {}).get('os', {}).get('family')
        if os_family is None:
            raise Exception('Node %s do not contain necessary context data' % host)
        return os_family

    def get_os_family_for_nodes(self, hosts: Iterable[str]) -> str:
        """
        Returns the detected operating system family for hosts.

        :return: Detected OS family, possible values: "debian", "rhel", "rhel8", "rhel9", "multiple", "unknown", "unsupported".
        """
        os_families = {self.get_os_family_for_node(host) for host in hosts}
        if len(os_families) > 1:
            return 'multiple'
        elif len(os_families) == 0:
            raise Exception('Cannot get os family for empty nodes list')
        else:
            return list(os_families)[0]

    def get_os_family(self) -> str:
        """
        Returns common OS family name from all final remote hosts.
        The method can be used during enrichment when NodeGroups are not yet calculated.

        :return: Detected OS family, possible values: "debian", "rhel", "rhel8", "rhel9", "multiple", "unknown", "unsupported".
        """
        hosts_detect_os_family = []
        for node in self.inventory['nodes']:
            host = self.get_access_address_from_node(node)
            if 'remove_node' not in node['roles']:
                hosts_detect_os_family.append(host)

        return self.get_os_family_for_nodes(hosts_detect_os_family)

    def get_os_identifiers(self) -> Dict[str, Tuple[str, str]]:
        """
        For each final node of the cluster, returns a tuple of OS (family, version).
        """
        nodes_check_os = self.nodes['all'].get_final_nodes()
        os_ids = {}
        for host in nodes_check_os.get_hosts():
            os_details = self.context['nodes'][host]['os']
            os_ids[host] = (os_details['family'], os_details['version'])

        return os_ids

    def _get_associations(self, os_family: str) -> Dict[str, dict]:
        if os_family in ('unknown', 'unsupported', 'multiple'):
            raise Exception("Failed to get associations for unsupported or multiple OS families")

        associations: dict = self.inventory['services']['packages']['associations'][os_family]
        return associations

    def get_associations(self) -> Dict[str, dict]:
        """
        Returns association for all packages from inventory for the cluster.
        The method can be used only if cluster has nodes with the same and supported OS family.
        """
        return self._get_associations(self.get_os_family())

    def _get_associations_for_os(self, os_family: str, package: str) -> dict:
        associations = self._get_associations(os_family).get(package)
        if associations is None:
            raise Exception(f'Failed to get associations for package "{package}"')

        return associations

    def get_associations_for_node(self, host: str, package: str) -> dict:
        """
        Returns all packages associations for specific node.

        :param host: The address of the node for which required to find the associations
        :param package: The package name to get the associations for
        :return: Dict with packages and their associations
        """
        node_os_family = self.get_os_family_for_node(host)
        return self._get_associations_for_os(node_os_family, package)

    def _get_package_associations_for_os(self, os_family: str, package: str, association_key: str) -> Any:
        associations = self._get_associations_for_os(os_family, package)
        association_value = associations.get(association_key)
        if association_value is None:
            raise Exception(f'Failed to get association "{association_key}" for package "{package}"')
        if not isinstance(association_value, str) and not isinstance(association_value, list):
            raise Exception(f'Unsupported association "{association_key}" value type for package "{package}", '
                            f'got: {str(association_value)}')

        return association_value

    def get_package_association(self, package: str, association_key: str) -> Any:
        """
        Returns the specified association for the specified package from inventory for the cluster.
        The method can be used only if cluster has nodes with the same and supported OS family.

        :param package: The package name to get the association for
        :param association_key: Association key to get
        :return: Association string or list value
        """
        os_family = self.get_os_family()
        return self._get_package_associations_for_os(os_family, package, association_key)

    def get_package_association_for_node(self, host: str, package: str, association_key: str) -> Any:
        """
        Returns the specified association for the specified package from inventory for specific node.

        :param host: The address of the node for which required to find the association
        :param package: The package name to get the association for
        :param association_key: Association key to get
        :return: Association string or list value
        """
        os_family = self.get_os_family_for_node(host)
        return self._get_package_associations_for_os(os_family, package, association_key)

    def make_finalized_inventory(self) -> dict:
        from kubemarine.core import defaults
        from kubemarine.procedures import add_node, remove_node
        from kubemarine import admission, controlplane, cri, packages

        cluster_finalized_functions: List[Callable[[KubernetesCluster, dict], dict]] = [
            packages.cache_package_versions,
            packages.remove_unused_os_family_associations,
            cri.remove_invalid_cri_config,
            add_node.add_node_finalize_inventory,
            remove_node.remove_node_finalize_inventory,
            admission.update_finalized_inventory,
            defaults.escape_jinja_characters_for_inventory,
            controlplane.controlplane_finalize_inventory,
        ]

        # copying is currently not necessary, but it is possible in general.
        prepared_inventory = self.inventory
        for finalize_fn in cluster_finalized_functions:
            prepared_inventory = finalize_fn(self, prepared_inventory)

        return defaults.prepare_for_dump(prepared_inventory, copy=False)

    def dump_finalized_inventory(self) -> None:
        inventory_for_dump = self.make_finalized_inventory()
        data = yaml.dump(inventory_for_dump)
        finalized_filename = "cluster_finalized.yaml"
        utils.dump_file(self, data, finalized_filename)
        utils.dump_file(self, data, finalized_filename, dump_location=False)

    def preserve_inventory(self) -> None:
        self.log.debug("Start preserving of the information about the procedure.")
        cluster_storage = utils.ClusterStorage(self)
        cluster_storage.make_dir()
        if self.context.get('initial_procedure') == 'add_node':
            cluster_storage.upload_info_new_control_planes()
        cluster_storage.collect_procedure_info()
        cluster_storage.compress_and_upload_archive()
        cluster_storage.rotation_file()



# Copyright 2021-2023 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from textwrap import dedent

from kubemarine.core.action import Action
from kubemarine.core.patch import RegularPatch
from kubemarine.core.resources import DynamicResources
from kubemarine import sysctl
from kubemarine.core import utils
import io

class TheAction(Action):
    def __init__(self) -> None:
        super().__init__("Set sysctl variables")

    def run(self, res: DynamicResources) -> None:
        cluster = res.cluster()
        node_group = cluster.make_group_from_roles(['all'])
        group_os_family = node_group.get_nodes_os()

        # add nf_conntrack module to predefined modules list
        # and load it without the node reboot
        for node in node_group.get_ordered_members_list():
            config = ''
            raw_config = ''

            for module_name in cluster.inventory['services']['modprobe'][node.get_nodes_os()]:
                module_name = module_name.strip()
                if module_name is not None and module_name != '':
                    config += module_name + "\n"
                    raw_config += module_name + " "

            dump_filename = 'modprobe_predefined.conf'
            if group_os_family == 'multiple':
                dump_filename = f'modprobe_predefined_{node.get_node_name()}.conf'

            utils.dump_file(cluster, config, dump_filename)
            node.put(io.StringIO(config), "/etc/modules-load.d/predefined.conf", backup=True, sudo=True)
            node.sudo("modprobe -a %s" % raw_config)
 

        # configure syslog variables
        node_group.call(sysctl.configure)
        node_group.call(sysctl.reload)
        

class SetSysctlVariables(RegularPatch):
    def __init__(self) -> None:
        super().__init__("set_sysctl_variables")

    @property
    def action(self) -> Action:
        return TheAction()

    @property
    def description(self) -> str:
        return dedent(
            f"""\
            This patch sets kernel variables with sysctl at all the nodes according to the new defaults.
            """.rstrip()
        )



#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import copy

from typing import Any, OrderedDict, List

from kubemarine import kubernetes, packages, plugins
from kubemarine.core import flow
from kubemarine.core.action import Action
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.resources import DynamicResources
from kubemarine.plugins.nginx_ingress import redeploy_ingress_nginx_is_needed
from kubemarine.procedures import install


def deploy_kubernetes_join(cluster: KubernetesCluster) -> None:

    group = cluster.make_group_from_roles(['control-plane', 'worker']).get_new_nodes()

    if group.is_empty():
        cluster.log.debug("No kubernetes nodes to perform")
        return

    cluster.nodes['control-plane'].get_new_nodes().call(kubernetes.join_new_control_plane)

    if "worker" in cluster.nodes:
        cluster.nodes['worker'].get_new_nodes().exclude_group(cluster.nodes['control-plane']) \
            .call(kubernetes.init_workers)

    group.call_batch([
        kubernetes.apply_labels,
        kubernetes.apply_taints
    ])

    cluster.log.debug("Waiting for new kubernetes nodes...")
    kubernetes.wait_for_nodes(group)
    kubernetes.schedule_running_nodes_report(cluster)


def redeploy_plugins_if_needed(cluster: KubernetesCluster) -> None:
    # redeploy ingress-nginx-controller if needed
    if redeploy_ingress_nginx_is_needed(cluster):
        cluster.log.debug("Redeploy ingress-nginx-controller plugin")
        plugins.install_plugin(cluster, 'nginx-ingress-controller',
                               cluster.inventory['plugins']['nginx-ingress-controller']['installation']['procedures'])
    else:
        cluster.log.debug("Redeploy ingress-nginx-controller is not needed, skip it")


def add_node_finalize_inventory(cluster: KubernetesCluster, inventory_to_finalize: dict) -> dict:
    if cluster.context.get('initial_procedure') != 'add_node':
        return inventory_to_finalize

    is_finalization = any('add_node' in node['roles'] for node in inventory_to_finalize['nodes'])

    if not is_finalization:
        # new nodes are not presented in final inventory - let's add it original config
        kubernetes.add_node_enrichment(inventory_to_finalize, cluster)

    # new nodes are already presented in final inventory - ok, just remove label
    for i, node in enumerate(inventory_to_finalize['nodes']):
        if 'add_node' in node['roles']:
            node['roles'].remove('add_node')

    return inventory_to_finalize


def cache_installed_packages(cluster: KubernetesCluster) -> None:
    """
    Task which is used to collect already installed packages versions on already existing nodes.
    It is called first during "add_node" procedure,
    so that new nodes install exactly the same packages as on other already existing nodes.
    """
    packages.cache_package_versions(cluster, cluster.inventory, by_initial_nodes=True)


tasks: OrderedDict[str, Any] = collections.OrderedDict(copy.deepcopy(install.tasks))
del tasks["deploy"]["accounts"]
tasks["deploy"]["plugins"] = redeploy_plugins_if_needed
tasks["deploy"]["kubernetes"]["init"] = deploy_kubernetes_join
tasks["cache_packages"] = cache_installed_packages
tasks.move_to_end("cache_packages", last=False)


class AddNodeAction(Action):
    def __init__(self) -> None:
        super().__init__('add node', recreate_inventory=True)

    def run(self, res: DynamicResources) -> None:
        flow.run_tasks(res, tasks, cumulative_points=install.cumulative_points)
        res.make_final_inventory()


def create_context(cli_arguments: List[str] = None) -> dict:

    cli_help = '''
    Script for adding node to Kubernetes cluster.

    How to use:

    '''

    parser = flow.new_procedure_parser(cli_help, tasks=tasks)
    context = flow.create_context(parser, cli_arguments, procedure='add_node')
    return context


def main(cli_arguments: List[str] = None) -> None:
    context = create_context(cli_arguments)
    flow.ActionsFlow([AddNodeAction()]).run_flow(context)


if __name__ == '__main__':
    main()



# Copyright 2021-2023 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from textwrap import dedent
from typing import Optional

from kubemarine.core.action import Action
from kubemarine.core.patch import RegularPatch
from kubemarine.core.resources import DynamicResources


class TheAction(Action):
    def __init__(self) -> None:
        super().__init__("Set previously resolved interfaces for VRRP IPs")

    def run(self, res: DynamicResources) -> None:
        logger = res.logger()
        cluster = res.cluster()

        if cluster.make_group_from_roles(['keepalived']).is_empty():
            return logger.info("Skipping the patch as Keepalived is not configured.")

        first_resolved_interface: Optional[str] = None
        for i, raw_item in enumerate(cluster.raw_inventory['vrrp_ips']):
            # If hosts are not specified, they are resolved by new algorithm.
            if isinstance(raw_item, str) or 'hosts' not in raw_item:
                item = cluster.inventory['vrrp_ips'][i]
                # There is at least one balancer that produced at least one host for VRRP IP
                resolved_interface = item['hosts'][0]['interface']
                if first_resolved_interface is None:
                    first_resolved_interface = resolved_interface
                    continue

                # The further VRRP IPs without hosts previously was enriched with 'first_resolved_interface'
                # Let's compare it with how it is resolved now.
                if resolved_interface != first_resolved_interface:
                    logger.info(f"Changing vrrp_ips[{i}].interface to {first_resolved_interface} "
                                f"to preserve the backward compatibility.")
                    formatted_item = res.formatted_inventory()['vrrp_ips'][i]
                    if isinstance(raw_item, str):
                        res.formatted_inventory()['vrrp_ips'][i] = formatted_item = {
                            'ip': formatted_item
                        }

                    formatted_item['interface'] = first_resolved_interface
                    item['interface'] = first_resolved_interface
                    for record in item['hosts']:
                        record['interface'] = first_resolved_interface

                    self.recreate_inventory = True

        if not self.recreate_inventory:
            logger.info("Skipping the patch as interfaces for `vrrp_ips` were not changed.")


class FixVRRP_IPsInterfaces(RegularPatch):
    def __init__(self) -> None:
        super().__init__("fix_vrrp_ips_interfaces")

    @property
    def action(self) -> Action:
        return TheAction()

    @property
    def description(self) -> str:
        return dedent(
            f"""\
            The patch sets previously resolved interfaces for the VRRP IPs
            in the inventory to preserve the backward compatibility.
            """.rstrip()
        )



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional

from distutils.util import strtobool
from kubemarine.core import static
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.group import RunnersGroupResult, NodeGroup
from kubemarine.cri import docker, containerd


def enrich_inventory(inventory: dict, cluster: KubernetesCluster) -> dict:
    cri_impl = inventory['services']['cri']['containerRuntime']
    if cluster.context.get("initial_procedure") != "migrate_cri":
        if cri_impl == "docker":
            forbidden_cri_sections = {"containerd": "containerdConfig"}
        else:
            forbidden_cri_sections = {"docker": "dockerConfig"}
        for key, value in forbidden_cri_sections.items():
            if value in cluster.raw_inventory.get('services', {}).get('cri', {}):
                raise Exception(f"{key} is not used, please remove {value} config from `services.cri` section")

    # Enrich containerdConfig
    if cri_impl == "containerd":
        return containerd.enrich_inventory(inventory, cluster)

    return inventory


def enrich_upgrade_inventory(inventory: dict, cluster: KubernetesCluster) -> dict:
    if cluster.context.get("initial_procedure") != "upgrade":
        return inventory

    cri_impl = inventory['services']['cri']['containerRuntime']
    if cri_impl == "containerd":
        return containerd.enrich_upgrade_inventory(inventory, cluster)

    return inventory


def upgrade_finalize_inventory(cluster: KubernetesCluster, inventory: dict) -> dict:
    if cluster.context.get("initial_procedure") != "upgrade":
        return inventory

    cri_impl = cluster.inventory['services']['cri']['containerRuntime']
    if cri_impl == "containerd":
        return containerd.upgrade_finalize_inventory(cluster, inventory)

    return inventory


def get_initial_cri_impl(inventory: dict) -> str:
    cri_impl: Optional[str] = inventory.get("services", {}).get("cri", {}).get("containerRuntime")
    if cri_impl is None:
        cri_impl = static.DEFAULTS['services']['cri']['containerRuntime']

    return cri_impl


def remove_invalid_cri_config(cluster: KubernetesCluster, inventory: dict) -> dict:
    if inventory['services']['cri']['containerRuntime'] == 'docker':
        if inventory['services']['cri'].get('containerdConfig'):
            del inventory['services']['cri']['containerdConfig']
    elif inventory['services']['cri'].get('dockerConfig'):
        del inventory['services']['cri']['dockerConfig']

    return inventory


def install(group: NodeGroup) -> RunnersGroupResult:
    cri_impl = group.cluster.inventory['services']['cri']['containerRuntime']

    if cri_impl == "docker":
        return docker.install(group)
    else:
        return containerd.install(group)


def configure(group: NodeGroup) -> RunnersGroupResult:
    cri_impl = group.cluster.inventory['services']['cri']['containerRuntime']

    if cri_impl == "docker":
        return docker.configure(group)
    else:
        return containerd.configure(group)


def prune(group: NodeGroup, all_implementations: bool = False) -> RunnersGroupResult:
    cri_impl = group.cluster.inventory['services']['cri']['containerRuntime']

    result = RunnersGroupResult(group.cluster, {})
    if cri_impl == "docker" or all_implementations:
        result = docker.prune(group)

    if cri_impl == "containerd" or all_implementations:
        result = containerd.prune(group)

    return result



#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
import ipaddress
import math
import os
import re
import sys
import uuid
from collections import OrderedDict
import time
from contextlib import contextmanager
from typing import List, Dict, Any, cast, Match, Iterator, Optional

import yaml

from kubemarine.core import flow, utils
from kubemarine import system, packages
from kubemarine.core.action import Action
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.resources import DynamicResources
from kubemarine.testsuite import TestSuite, TestCase, TestFailure, TestWarn
from kubemarine.core.group import NodeConfig, GroupException, GroupResultException, CollectorCallback


def connection_ssh_connectivity(cluster: KubernetesCluster) -> None:
    with TestCase(cluster, '001', 'SSH', 'Connectivity', default_results='Connected'):
        failed_nodes = []
        for node in cluster.nodes['all'].get_ordered_members_list():
            try:
                cluster.log.verbose(node.run("echo 1"))
            except GroupException as e:
                failed_nodes.append(node.get_node_name())
                cluster.log.error("Connection test failed for node \"%s\"" % node.get_node_name())
                cluster.log.error("Exception details:")
                cluster.log.error(e)
        if failed_nodes:
            raise TestFailure("Failed to connect to %s nodes" % len(failed_nodes),
                              hint="Failed to connect from the deploy node to the remote node of the cluster. Check that "
                                   "the inventory details (key, username, and nodes addresses) are entered correctly, and verify "
                                   "the access to remote nodes.")


def connection_ssh_latency_single(cluster: KubernetesCluster) -> None:
    with TestCase(cluster, '002',  'SSH', 'Latency - Single Thread',
                  minimal=cluster.globals['compatibility_map']['network']['connection']['latency']['single']['critical'],
                  recommended=cluster.globals['compatibility_map']['network']['connection']['latency']['single']['recommended']) as tc:
        i = 0
        measurements = []
        while i < 5:
            i += 1
            for node in cluster.nodes['all'].get_ordered_members_list():
                time_start = time.time()
                node.run("echo 1")
                time_end = time.time()
                diff = (time_end - time_start) * 1000
                cluster.log.debug('Connection to %s - %sms' % (node.get_node_name(), diff))
                measurements.append(diff)
        average_latency = math.floor(sum(measurements) / cluster.nodes['all'].nodes_amount() / 5)
        if average_latency > cluster.globals['compatibility_map']['network']['connection']['latency']['single']['critical']:
            raise TestFailure("Very high latency: %sms" % average_latency,
                              hint="A very high latency was detected between the deploy node and cluster nodes. "
                                   "Check your network settings and status. It is necessary to reduce the latency to %sms."
                                   % cluster.globals['compatibility_map']['network']['connection']['latency']['single']['critical'])
        if average_latency > cluster.globals['compatibility_map']['network']['connection']['latency']['single']['recommended']:
            raise TestWarn("High latency: %sms" % average_latency,
                           hint="The detected latency is higher than the recommended value (%sms). Check your network settings "
                                "and status." % cluster.globals['compatibility_map']['network']['connection']['latency']['single']['recommended'])
        tc.success(results="%sms" % average_latency)


def connection_ssh_latency_multiple(cluster: KubernetesCluster) -> None:
    with TestCase(cluster, '003',  'SSH', 'Latency - Multi Thread',
                  minimal=cluster.globals['compatibility_map']['network']['connection']['latency']['multi']['critical'],
                  recommended=cluster.globals['compatibility_map']['network']['connection']['latency']['multi']['recommended']) as tc:
        i = 0
        measurements = []
        while i < 10:
            i += 1
            time_start = time.time()
            cluster.nodes['all'].run("echo 1")
            time_end = time.time()
            diff = (time_end - time_start) * 1000
            cluster.log.debug('Average latency at step %s - %sms' % (i, diff))
            measurements.append(diff)
        average_latency = math.floor(sum(measurements) / 10)
        if average_latency > cluster.globals['compatibility_map']['network']['connection']['latency']['multi']['critical']:
            raise TestFailure("Very high latency: %sms" % average_latency,
                              hint="A very high latency was detected between the deploy node and cluster nodes. "
                                   "Check your network settings and status. It is necessary to reduce the latency to %sms."
                                   % cluster.globals['compatibility_map']['network']['connection']['latency']['multi']['critical'])
        if average_latency > cluster.globals['compatibility_map']['network']['connection']['latency']['multi']['recommended']:
            raise TestWarn("High latency: %sms" % average_latency,
                           hint="The detected latency is higher than the recommended value (%sms). Check your network settings "
                                "and status." % cluster.globals['compatibility_map']['network']['connection']['latency']['multi']['recommended'])
        tc.success(results="%sms" % average_latency)


def connection_sudoer_access(cluster: KubernetesCluster) -> None:
    with TestCase(cluster, '004', 'SSH', 'Sudoer Access', default_results='Access provided'):
        non_root = []
        for host, node_context in cluster.context['nodes'].items():
            access_info = node_context['access']
            if access_info['online'] and access_info['sudo'] == 'Root':
                cluster.log.debug("%s online and has root" % host)
            else:
                non_root.append(host)
        if non_root:
            raise TestFailure("Non-sudoer access found at: %s" % ", ".join(non_root),
                              hint="Certain nodes do not have the appropriate sudoer access. At these nodes, add "
                                   "a connection user to the sudoers group.")


def hardware_members_amount(cluster: KubernetesCluster, group_name: str) -> None:
    beauty_name = group_name.capitalize()
    if group_name == 'vip':
        beauty_name = 'VIP'
    if group_name == 'all':
        beauty_name = 'Total Node'

    with TestCase(cluster, '005',  'Hardware', '%ss Amount' % beauty_name,
                  minimal=cluster.globals['compatibility_map']['hardware']['minimal'][group_name]['amount'],
                  recommended=cluster.globals['compatibility_map']['hardware']['recommended'][group_name]['amount']) as tc:
        amount = 0
        if group_name == 'vip':
            amount = len(cluster.inventory.get('vrrp_ips', []))
        else:
            group = cluster.nodes.get(group_name)
            if group is not None:
                amount = group.nodes_amount()

        s = ''
        if amount != 1:
            s = 's'

        if amount < cluster.globals['compatibility_map']['hardware']['minimal'][group_name]['amount']:
            beauty_name = group_name
            if group_name == 'all':
                beauty_name = 'all node'
            raise TestFailure("Less than minimal. Detected %s item%s" % (amount, s),
                              hint="Increase the number of resources, so that the number of %ss in the cluster should not "
                                   "be less than %s." % (beauty_name, cluster.globals['compatibility_map']['hardware']['minimal'][group_name]['amount']))

        if amount < cluster.globals['compatibility_map']['hardware']['recommended'][group_name]['amount']:
            beauty_name = group_name
            if group_name == 'all':
                beauty_name = 'all node'
            raise TestWarn("Less than recommended. Detected %s item%s" % (amount, s),
                           hint="Increase the number of resources, so that the number of %ss in the cluster should not "
                                "be less than %s." % (beauty_name, cluster.globals['compatibility_map']['hardware']['minimal'][group_name]['amount']))

        tc.success("%s item%s" % (amount, s))


def hardware_cpu(cluster: KubernetesCluster, group_name: str) -> None:
    minimal_cpu = cluster.globals['compatibility_map']['hardware']['minimal'][group_name]['vcpu'] \
        if group_name == 'balancer' or cluster.nodes['all'].nodes_amount() > 1 \
        else cluster.globals['compatibility_map']['hardware']['minimal']['control-plane']['vcpu']
    with TestCase(cluster, '006',  'Hardware', 'VCPUs Amount - %ss' % group_name.capitalize(),
                  minimal=minimal_cpu,
                  recommended=cluster.globals['compatibility_map']['hardware']['recommended'][group_name]['vcpu']) as tc:
        if cluster.nodes.get(group_name) is None or cluster.nodes[group_name].is_empty():
            return tc.success(results='Skipped')
        results = cluster.nodes[group_name].sudo("nproc --all")
        cluster.log.verbose(results)
        minimal_amount: Optional[int] = None
        for host, result in results.items():
            amount = int(result.stdout)
            if minimal_amount is None or minimal_amount > amount:
                minimal_amount = amount
            if amount < minimal_cpu:
                cluster.log.error('%s node %s has insufficient VCPUs: expected %s, but %s found.'
                                  % (group_name.capitalize(), host, cluster.globals['compatibility_map']['hardware']['minimal'][group_name]['vcpu'], amount))
            elif amount < cluster.globals['compatibility_map']['hardware']['recommended'][group_name]['vcpu']:
                cluster.log.warning('%s node %s has less VCPUs than recommended: recommended %s, but %s found.'
                                    % (group_name.capitalize(), host, cluster.globals['compatibility_map']['hardware']['recommended'][group_name]['vcpu'], amount))
            else:
                cluster.log.debug('%s node %s has enough VCPUs: %s' % (group_name.capitalize(), host, amount))

        s = ''
        if minimal_amount != 1:
            s = 's'

        if minimal_amount < minimal_cpu:
            raise TestFailure("Less than minimal. Detected %s VCPU%s" % (minimal_amount, s),
                              hint="Increase the number of VCPUs in the node configuration to at least the minimum "
                                   "value: %s VCPUs." % cluster.globals['compatibility_map']['hardware']['minimal'][group_name]['vcpu'])
        if minimal_amount < cluster.globals['compatibility_map']['hardware']['recommended'][group_name]['vcpu']:
            raise TestWarn("Less than recommended. Detected %s VCPU%s" % (minimal_amount, s),
                           hint="Increase the number of VCPUs in the node configuration up to %s VCPUs."
                                % cluster.globals['compatibility_map']['hardware']['recommended'][group_name]['vcpu'])
        tc.success(results='%s VCPU%s' % (minimal_amount, s))


def hardware_ram(cluster: KubernetesCluster, group_name: str) -> None:
    with TestCase(cluster, '007',  'Hardware', 'RAM Amount - %ss' % group_name.capitalize(),
                  minimal=cluster.globals['compatibility_map']['hardware']['minimal'][group_name]['ram'],
                  recommended=cluster.globals['compatibility_map']['hardware']['recommended'][group_name]['ram']) as tc:
        if cluster.nodes.get(group_name) is None or cluster.nodes[group_name].is_empty():
            return tc.success(results='Skipped')
        results = cluster.nodes[group_name].sudo("cat /proc/meminfo | awk '/DirectMap/ { print $2 }'")
        cluster.log.verbose(results)
        minimal_amount: Optional[int] = None
        for host, result in results.items():
            amount = math.floor(sum(map(lambda x: int(x), result.stdout.strip().split("\n"))) / 1000000)
            if minimal_amount is None or minimal_amount > amount:
                minimal_amount = amount
            if amount < cluster.globals['compatibility_map']['hardware']['minimal'][group_name]['ram']:
                cluster.log.error('%s node %s has insufficient RAM: expected %sGB, but %sGB found.'
                                  % (group_name.capitalize(), host, cluster.globals['compatibility_map']['hardware']['minimal'][group_name]['ram'], amount))
            elif amount < cluster.globals['compatibility_map']['hardware']['recommended'][group_name]['ram']:
                cluster.log.warning('%s node %s has less RAM than recommended: recommended %sGB, but %sGB found.'
                                    % (group_name.capitalize(), host, cluster.globals['compatibility_map']['hardware']['recommended'][group_name]['ram'], amount))
            else:
                cluster.log.debug('%s node %s has enough RAM: %sGB' % (group_name.capitalize(), host, amount))
        if minimal_amount < cluster.globals['compatibility_map']['hardware']['minimal'][group_name]['ram']:
            raise TestFailure("Less than minimal. Detected %sGB" % minimal_amount,
                              hint="Increase the number of RAM in the node configuration to at least the minimum "
                                   "value: %sGB." % cluster.globals['compatibility_map']['hardware']['minimal'][group_name]['ram'])
        if minimal_amount < cluster.globals['compatibility_map']['hardware']['recommended'][group_name]['ram']:
            raise TestWarn("Less than recommended. Detected %sGB" % minimal_amount,
                           hint="Increase the number of RAM in the node configuration up to %s GB."
                                % cluster.globals['compatibility_map']['hardware']['recommended'][group_name]['ram'])
        tc.success(results='%sGB' % minimal_amount)


def system_distributive(cluster: KubernetesCluster) -> None:
    with TestCase(cluster, '008', 'System', 'Distibutive') as tc:
        supported_distributives = cluster.globals['compatibility_map']['distributives'].keys()

        cluster.log.debug(system.fetch_os_versions(cluster))

        detected_unsupported_os = []
        detected_supported_os = []
        detected_unsupported_version = []
        supported_versions = []
        for address, context_data in cluster.context["nodes"].items():
            detected_os = '%s %s' % (context_data['os']['name'], context_data['os']['version'])
            if context_data['os']['family'] == 'unsupported': 
                detected_unsupported_os.append(detected_os)
                cluster.log.error('Host %s running unsupported OS \"%s\"' % (address, detected_os))
            elif context_data['os']['family'] == 'unknown':
                detected_unsupported_version.append(detected_os)
                os_family_list = cluster.globals["compatibility_map"]["distributives"][context_data['os']['name']]
                versions = []
                for os_family_item in os_family_list:
                    versions.extend(os_family_item["versions"])
                supported_versions.append('%s: %s' %(context_data['os']['name'], versions))
                cluster.log.error('Host %s running unknown OS family \"%s\"' % (address, detected_os))
            else:
                detected_supported_os.append(detected_os)
                cluster.log.debug('Host %s running \"%s\"' % (address, detected_os))

        detected_supported_os = list(set(detected_supported_os))
        detected_unsupported_os = list(set(detected_unsupported_os))
        detected_unsupported_version = list(set(detected_unsupported_version))
        supported_versions = list(set(supported_versions))

        if detected_unsupported_os:
            raise TestFailure("Unsupported OS: %s" % ", ".join(detected_unsupported_os),
                              hint="Reinstall the OS on the host to one of the supported: %s" % ", ".join(supported_distributives))

        if detected_unsupported_version:
            raise TestFailure("Unsupported version: %s" % ", ".join(detected_unsupported_version),
                              hint="Reinstall the OS on the host to one of the supported versions: %s" % \
                                      ", ".join(supported_versions))
        
        os_ids = cluster.get_os_identifiers()
        different_os = set(os_ids.values())
        if len(different_os) > 1:
            cluster.log.warning(
                f"Nodes have different OS families or versions. "
                f"List of (OS family, version): {list(different_os)}")
            raise TestWarn(f"Nodes have different OS families or versions")
        
        tc.success(results=", ".join(detected_supported_os))


def check_kernel_version(cluster: KubernetesCluster) -> None:
    """
    This method compares the linux kernel version with the bad version
    """
    with TestCase(cluster, '015', "Software", "Kernel version") as tc:
        bad_results = {}
        unstable_kernel_ubuntu: List[str] = cluster.globals['compatibility_map']['distributives']['ubuntu'][0].get('unstable_kernel')
        unstable_kernel_centos: List[str] = []
        group = cluster.nodes['all']
        result_group = group.run('uname -r')
        for host, results in result_group.items():
            os_name = cluster.context['nodes'][host]['os']['name']
            result = results.stdout.rstrip()
            if os_name == 'ubuntu':
                if result in unstable_kernel_ubuntu:
                    bad_results[host] = result
            elif os_name == 'centos':
                if result in unstable_kernel_centos:
                    bad_results[host] = result

        if len(bad_results) > 0:
            for host, kernel_version in bad_results.items():
                cluster.log.debug(f"Unstable kernel %s on: %s" % (kernel_version, host))
            cluster.log.debug(f"Update the linux kernel version to 5.4.0-135-generic")
            raise TestWarn("Kernel version unstable")
        else:
            tc.success("All kernel have stable versions")


def check_access_to_thirdparties(cluster: KubernetesCluster) -> None:
    with TestCase(cluster, '012', 'Software', 'Thirdparties Availability') as tc:
        detect_preinstalled_python(cluster)
        broken = []
        nodes_without_python = set()

        # Load script for checking sources
        all_group = cluster.nodes['all']
        check_script = utils.read_internal("resources/scripts/check_url_availability.py")
        random_temp_path = "/tmp/%s.py" % uuid.uuid4().hex
        all_group.put(io.StringIO(check_script), random_temp_path)

        for destination, config in cluster.inventory['services'].get('thirdparties', {}).items():
            # Check if curl
            if config['source'][:4] != 'http' or '://' not in config['source'][4:8]:
                continue
            # Check with script
            common_group = cluster.create_group_from_groups_nodes_names(config.get('groups', []), config.get('nodes', []))
            for node in common_group.get_ordered_members_list():
                host = node.get_host()
                if cluster.context['nodes'][host]['python'] == "Not installed":
                    nodes_without_python.add(host)
                    continue
                python_executable = cluster.context['nodes'][host]['python']['executable']
                res = node.run("%s %s %s %s" % (python_executable, random_temp_path, config['source'],
                                                cluster.inventory['globals']['timeout_download']), warn=True)
                if res.is_any_failed():
                    broken.append(f"{host}, {destination}: {res[host].stderr}")

        # Remove file
        rm_command = "rm %s" % random_temp_path
        all_group.run(rm_command)

        if broken:
            raise TestFailure('Required thirdparties are unavailable', hint=yaml.safe_dump(broken))
        if nodes_without_python:
            raise TestWarn("Can't detect python version for some nodes, procedure can't be performed for them",
                           hint=yaml.safe_dump(list(nodes_without_python)))
        tc.success('All thirdparties are available')


def check_resolv_conf(cluster: KubernetesCluster) -> None:
    nodes_context = cluster.context['nodes']
    hosts = [host for host, node_context in nodes_context.items() if 'resolv_conf_is_actual' not in node_context]
    group = cluster.make_group(hosts)

    if cluster.inventory["services"].get("resolv.conf") is None:
        for host in hosts:
            nodes_context[host]["resolv_conf_is_actual"] = True
    else:
        # Create temp resolv.conf file
        resolv_conf_buffer = system.get_resolv_conf_buffer(cluster.inventory["services"].get("resolv.conf"))
        random_resolv_conf_path = "/tmp/%s.conf" % uuid.uuid4().hex
        group.put(resolv_conf_buffer, random_resolv_conf_path)

        # Compare with existed resolv.conf
        results = group.run(
            '[ -f /etc/resolv.conf ] && '
            'cmp --silent /etc/resolv.conf %s' % random_resolv_conf_path,
            warn=True)

        for host, res in results.items():
            nodes_context[host]["resolv_conf_is_actual"] = not res.failed
        # Remove temp resolv.conf file
        group.run("rm %s" % random_resolv_conf_path)


def check_package_repositories(cluster: KubernetesCluster) -> None:
    nodes_context = cluster.context['nodes']
    hosts = [host for host, node_context in nodes_context.items() if 'package_repos_are_actual' not in node_context]

    repositories = cluster.inventory['services']['packages']['package_manager'].get("repositories")
    if repositories is None:
        for host in hosts:
            nodes_context[host]["package_repos_are_actual"] = True
    else:
        group = cluster.make_group(hosts)
        random_repos_conf_path = "/tmp/%s.repo" % uuid.uuid4().hex
        collector = CollectorCallback(cluster)
        with group.new_executor() as exe:
            for node in exe.group.get_ordered_members_list():
                # Create temp repos file
                packages.create_repo_file(node, repositories, random_repos_conf_path)

                # Compare with existed resolv.conf
                predefined_repos_file = packages.get_repo_filename(node)
                node.sudo(
                    '[ -f %s ] && cmp --silent %s %s' %
                    (predefined_repos_file, predefined_repos_file, random_repos_conf_path),
                    warn=True, callback=collector)

        for host, result in collector.result.items():
            nodes_context[host]["package_repos_are_actual"] = not result.failed

        # Remove temp .repo file
        group.sudo("rm %s" % random_repos_conf_path)


def check_access_to_package_repositories(cluster: KubernetesCluster) -> None:
    with TestCase(cluster, '013', 'Software', 'Package Repositories') as tc:
        detect_preinstalled_python(cluster)
        check_resolv_conf(cluster)
        broken = []
        warnings = []

        # Collect repository urls
        # TODO: think about better parsing
        repository_urls: List[str] = []
        repositories = cluster.inventory['services']['packages']['package_manager'].get("repositories")
        if cluster.get_os_family() not in ['debian', 'rhel', 'rhel8', 'rhel9']:
            # Skip check in case of multiply or unknown OS
            raise TestWarn("Can't check package repositories on multiply OS")
        if isinstance(repositories, list):
            # For debian
            for repo in repositories:
                repository_url = list(filter(lambda x: x[:4] == 'http' and '://' in x[4:8], repo.split(' ')))
                if repository_url:
                    repository_urls.append(repository_url[0])
                else:
                    broken.append(f"Found broken repository: '{repo}'")
        elif isinstance(repositories, dict):
            # For rhel
            for repo_name, repo_conf in repositories.items():
                if repo_conf.get('baseurl') is not None:
                    repository_urls.append(repo_conf.get('baseurl'))
                else:
                    broken.append(f"Found broken repository: '{repo_name}'")
        elif isinstance(repositories, str):
            path = utils.get_external_resource_path(repositories)
            if not os.path.isfile(path):
                broken.append(f"File {path} with the repositories content does not exist")
            else:
                repositories = utils.read_external(path)
                for repo in repositories.split('\n'):
                    repository_url = list(filter(lambda x: x[:4] == 'http' and '://' in x[4:8], repo.split(' ')))
                    if repository_url:
                        repository_urls.append(repository_url[0])
                if not repository_urls:
                    broken.append(f"Failed to detect repository URLs in file {path}")

        repository_urls = list(set(repository_urls))
        cluster.log.debug(f"Repositories to check: {repository_urls}")

        # Load script for checking sources
        all_group = cluster.nodes['all']
        check_script = utils.read_internal("resources/scripts/check_url_availability.py")
        random_temp_path = "/tmp/%s.py" % uuid.uuid4().hex
        all_group.put(io.StringIO(check_script), random_temp_path)

        if repository_urls:
            collector = CollectorCallback(cluster)
            with all_group.new_executor() as exe:
                for node in exe.group.get_ordered_members_list():
                    host = node.get_host()
                    # Check with script
                    if cluster.context['nodes'][host]['python'] == 'Not installed':
                        warnings.append(f"Can't detect python version for node {host}, "
                                        f"operation can't be performed for it")
                        continue
                    python_executable = cluster.context['nodes'][host]['python']['executable']
                    for repo_url in repository_urls:
                        node.run('%s %s %s %s'
                                 % (python_executable, random_temp_path, repo_url,
                                    cluster.inventory['globals']['timeout_download']),
                                 warn=True, callback=collector)

            for host, url_results in collector.results.items():
                # Check if resolv.conf is actual
                resolv_conf_actual = cluster.context['nodes'][host]['resolv_conf_is_actual']
                if not resolv_conf_actual:
                    warnings.append(f"resolv.conf is not installed for node {host}: "
                                    f"Package repositories can be unavailable. You can install resolv.conf using task "
                                    f"`install --tasks prepare.dns.resolv_conf`")
                    problem_handler = warnings
                else:
                    problem_handler = broken
                for i, result in enumerate(url_results):
                    if result.failed:
                        problem_handler.append(f"{host}, {repository_urls[i]}: {result.stderr}")

        # Remove file
        rm_command = "rm %s" % random_temp_path
        all_group.run(rm_command)

        if broken:
            raise TestFailure('Found problems for package repositories', hint=yaml.safe_dump(broken))
        elif warnings:
            raise TestWarn('Found potential problems for package repositories or nodes', hint=yaml.safe_dump(warnings))
        tc.success('All package repositories are correct and available')


def check_access_to_packages(cluster: KubernetesCluster) -> None:
    with TestCase(cluster, '014', 'Software', 'Package Availability') as tc:
        check_package_repositories(cluster)
        broken: List[str] = []
        warnings: List[str] = []
        group = cluster.nodes['all']
        hosts_to_packages = packages.get_all_managed_packages_for_group(group, cluster.inventory)
        collector = CollectorCallback(cluster)
        with group.new_executor() as exe:
            for node in exe.group.get_ordered_members_list():
                host = node.get_host()
                packages_to_check = list(set(hosts_to_packages[host]))
                hosts_to_packages[host] = packages_to_check
                cluster.log.debug(f"Packages to check for node {host}: {packages_to_check}")

                for package in packages_to_check:
                    packages.search_package(node, package, callback=collector)

        # Check packages from install section
        for host, results in collector.results.items():
            package_repos_are_actual = cluster.context['nodes'][host]["package_repos_are_actual"]
            if not package_repos_are_actual:
                warnings.append(f"Package repositories are not installed for {host}: "
                                f"Packages can be unavailable. You can install it using tasks "
                                f"`install --tasks prepare.dns.resolv_conf,prepare.package_manager.configure`")
                problem_handler = warnings
            else:
                problem_handler = broken
            packages_to_check = hosts_to_packages[host]
            for i, result in enumerate(results):
                if result.failed:
                    problem_handler.append(f"Package {packages_to_check[i]} is unavailable for node {host}")

        if broken:
            raise TestFailure('Required packages are unavailable', hint=yaml.safe_dump(broken))
        elif warnings:
            raise TestWarn('Found potential problems for packages', hint=yaml.safe_dump(warnings))
        tc.success('All packages are available')


def detect_preinstalled_python(cluster: KubernetesCluster) -> None:
    version_pattern = r'^Python{space}([2-3])(\.[0-9]+){{0,2}}$'
    bash_version_pattern = version_pattern.format(space='[[:space:]]')
    python_version_pattern = version_pattern.format(space=' ')
    nodes_context = cluster.context['nodes']
    hosts_unknown_python = [host for host, node_context in nodes_context.items() if 'python' not in node_context]
    group_unknown_python = cluster.make_group(hosts_unknown_python)
    detected_python = group_unknown_python.run(
        rf'for i in $(whereis -b python && whereis -b python3 ); do '
        rf'if [[ -f "$i" ]] && [[ $($i --version 2>&1 | head -n 1) =~ {bash_version_pattern} ]]; then '
        rf'echo "$i"; $i --version 2>&1; break; '
        rf'fi; done')

    for host, result in detected_python.items():
        identity = result.stdout.strip()
        if not identity:
            nodes_context[host]["python"] = "Not installed"
        else:
            executable, version = tuple(identity.splitlines())
            version = cast(Match[str], re.match(python_version_pattern, version)).group(1)
            nodes_context[host]["python"] = {
                "executable": executable,
                "major_version": version
            }


@contextmanager
def suspend_firewalld(cluster: KubernetesCluster) -> Iterator[None]:
    firewalld_statuses = system.fetch_firewalld_status(cluster.nodes["all"])
    stop_firewalld_group = firewalld_statuses.get_nodes_group_where_value_in_stdout("active (running)")

    nodes_to_rollback = cluster.make_group([])
    try:
        try:
            nodes_to_rollback = system.stop_service(stop_firewalld_group, "firewalld").get_group()
        except GroupException as e:
            nodes_to_rollback = e.get_exited_nodes_group()
            raise

        yield
    finally:
        system.start_service(nodes_to_rollback, "firewalld")


def _get_not_balancers(cluster: KubernetesCluster) -> Dict[str, NodeConfig]:
    nodes = {}
    for node in cluster.inventory['nodes']:
        # exclude nodes which are only balancers.
        if node["roles"] == ["balancer"]:
            cluster.log.debug(f"Exclude balancer '{node['name']}' from subnet connectivity check.")
            continue
        nodes[node["connect_to"]] = node

    return nodes


@contextmanager
def assign_random_ips(cluster: KubernetesCluster, nodes: Dict[str, NodeConfig], subnet: str) -> Iterator[Dict[str, Any]]:
    inet = ipaddress.ip_network(subnet)
    net_mask = str(inet.netmask)
    prefix = str(inet.prefixlen)
    subnet_hosts = []
    ip_numbers = 0
    for addr in inet.hosts():
        subnet_hosts.append(addr)
        ip_numbers += 1
        if ip_numbers == 1000000:
           break
    subnet_hosts_len = len(subnet_hosts)

    host_to_inf = {}
    host_to_ip = {}
    skipped_nodes = []
    nodes_to_rollback = cluster.make_group([])

    try:
        # Assign random IP for the subnet on every node
        i = 30
        for host, node_config in nodes.items():
            inf = cluster.context['nodes'][host]['active_interface']
            if not inf:
                raise TestFailure(f"Failed to detect active interface on {node_config['name']}")
            host_to_inf[host] = inf
            random_host = subnet_hosts[subnet_hosts_len - i]
            host_to_ip[host] = random_host
            i = i + 1

        collector = CollectorCallback(cluster)
        with cluster.make_group(nodes).new_executor() as exe:
            for node in exe.group.get_ordered_members_list():
                host = node.get_host()
                existing_alias = f"ip -o a | grep {host_to_inf[host]} | grep {host_to_ip[host]}"
                node.sudo(existing_alias, warn=True, callback=collector)

        for host, result in collector.result.items():
            if not result.stdout and not result.stderr and result.exited == 1:
                # grep returned nothing, subnet is not used.
                pass
            else:
                skipped_nodes.append(nodes[host]["name"])
                del nodes[host]

        try:
            with cluster.make_group(nodes).new_executor() as exe:
                # Create alias from the node network interface for the subnet on every node
                for node in exe.group.get_ordered_members_list():
                    host = node.get_host()
                    node.sudo(f"ip a add {host_to_ip[host]}/{prefix} dev {host_to_inf[host]}")

            nodes_to_rollback = cluster.make_group(nodes.keys())
        except GroupException as e:
            nodes_to_rollback = e.get_exited_nodes_group()
            raise

        yield host_to_ip
    finally:
        # Remove the created aliases from network interfaces
        with nodes_to_rollback.new_executor() as exe:
            for node in exe.group.get_ordered_members_list():
                host = node.get_host()
                node.sudo(f"ip a del {host_to_ip[host]}/{prefix} dev {host_to_inf[host]}",
                          warn=True)

    if skipped_nodes:
        raise TestWarn(f"Cannot perform check on {skipped_nodes}: subnet is already in use. "
                       f"Use check_paas procedure if you already have installed cluster.")


def pod_subnet_connectivity(cluster: KubernetesCluster) -> None:
    with TestCase(cluster, '009', 'Network', 'PodSubnet', default_results='Connected'),\
            suspend_firewalld(cluster):
        pod_subnet = cluster.inventory['services']['kubeadm']['networking']['podSubnet']
        nodes = _get_not_balancers(cluster)
        tcp_ports = ["30050"]
        with assign_random_ips(cluster, nodes, pod_subnet) as host_to_ip, \
                install_tcp_listener(cluster, nodes, tcp_ports):
            failed_nodes = check_tcp_connect_between_all_nodes(cluster, list(nodes.values()), tcp_ports, host_to_ip)

            if failed_nodes:
                raise TestFailure(f"Failed to connect to {len(failed_nodes)} nodes.",
                                  hint=f"Traffic is not allowed for the pod subnet({pod_subnet}) "
                                       f"on nodes: {failed_nodes}.")


def service_subnet_connectivity(cluster: KubernetesCluster) -> None:
    with TestCase(cluster, '010', 'Network', 'ServiceSubnet', default_results='Connected'),\
            suspend_firewalld(cluster):
        service_subnet = cluster.inventory['services']['kubeadm']['networking']['serviceSubnet']
        nodes = _get_not_balancers(cluster)
        tcp_ports = ["30050"]
        with assign_random_ips(cluster, nodes, service_subnet) as host_to_ip, \
                install_tcp_listener(cluster, nodes, tcp_ports):
            failed_nodes = check_tcp_connect_between_all_nodes(cluster, list(nodes.values()), tcp_ports, host_to_ip)

            if failed_nodes:
                raise TestFailure(f"Failed to connect to {len(failed_nodes)} nodes.",
                                  hint=f"Traffic is not allowed for the service subnet({service_subnet}) "
                                       f"on nodes: {failed_nodes}.")


def cmd_for_ports(ports: List[str], query: str) -> str:
    result = ""
    for port in ports:
        result += f" && echo 'port: {port}' && ( {query % port} ) "
    return result[3:]


def tcp_connect(cluster: KubernetesCluster, node_from: NodeConfig, node_to: NodeConfig,
                tcp_ports: List[str], host_to_ip: Dict[str, Any], mtu: int) -> None:
    # 40 bites for headers
    mtu -= 40
    cluster.log.verbose(f"Trying connection from '{node_from['name']}' to '{node_to['name']}")
    cmd = cmd_for_ports(tcp_ports, f"echo $(dd if=/dev/urandom bs={mtu}  count=1) >/dev/tcp/{host_to_ip[node_to['connect_to']]}/%s")
    group = cluster.make_group([node_from['connect_to']])
    group.sudo(cmd, timeout=cluster.globals['connection']['defaults']['timeout'])


def get_start_tcp_listener_cmd(python_executable: str, tcp_listener: str, ip_version: int) -> str:
    # 1. Create anonymous pipe
    # 2. Create python tcp listener process in background and redirect output to pipe
    # 3. Wait till the listener successfully binds the port, or till it fails and exits.
    #    Read one line from pipe to check that.
    # 4. Exit with success or fail correspondingly.
    return "PORT=%s; PIPE=$(mktemp -u); mkfifo $PIPE; exec 3<>$PIPE; rm $PIPE; " \
           f"sudo nohup {python_executable} {tcp_listener} $PORT {ip_version} >&3 2>&1 & " \
           "PID=$(echo $!); " \
           "while read -t 0.1 -u 3 || sudo kill -0 $PID 2>/dev/null && [[ -z $REPLY ]]; do " \
               ":; " \
           "done; " \
           "DATA=$REPLY; " \
           "if [[ $DATA == \"In use\" ]]; then " \
               "echo \"$PORT in use\" >&2 ; " \
               "exit 1; " \
           "elif [[ $DATA == \"Listen\" ]]; then " \
               "exit 0; " \
           "fi; " \
           "DATA=$(echo $DATA && dd iflag=nonblock status=none <&3 2>/dev/null); " \
           "echo \"$DATA\" >&2 ; " \
           "exit 1"


def get_stop_tcp_listener_cmd(tcp_listener: str) -> str:
    identify_pid = "ps aux | grep \" %s ${port} \" | grep -v grep | grep -v nohup | awk '{print $2}'" % tcp_listener
    return f"port=%s;pid=$({identify_pid}) " \
           "&& if [ ! -z $pid ]; then sudo kill -9 $pid; echo \"killed pid $pid for port $port\"; fi"


def check_tcp_connect_between_all_nodes(cluster: KubernetesCluster, node_list: List[NodeConfig],
                                        tcp_ports: List[str], host_to_ip: Dict[str, Any]) -> List[str]:
    if len(node_list) <= 1:
        return []

    mtu = cluster.inventory['plugins']['calico']['mtu']

    cluster.log.verbose("Searching for success node...")
    success_node = None
    failed_nodes = []
    for node in node_list:
        failed_nodes.append(node['name'])
    nodes_for_check = []
    for node in node_list:
        nodes_for_check.append(node)

    for i in range(0, len(node_list)):
        for j in range(i + 1, len(node_list)):
            try:
                tcp_connect(cluster, node_list[j], node_list[i], tcp_ports, host_to_ip, mtu)
                # If node has at least one successful connection with another node - this node has appropriate settings.
                success_node = node_list[i]
                cluster.log.verbose(f"Successful node found: {success_node['name']}")
                failed_nodes.remove(success_node["name"])
                break
            except Exception as e:
                cluster.log.error(f"Subnet connectivity test failed from '{node_list[j]['name']}' to '{node_list[i]['name']}'")
                cluster.log.verbose(f"Exception details: {e}")

        nodes_for_check.remove(node_list[i])
        if success_node is not None:
            break

    # TCP connect from found successful node to every other node
    if success_node is not None:
        for node in nodes_for_check:
            try:
                tcp_connect(cluster, success_node, node, tcp_ports, host_to_ip, mtu)
                failed_nodes.remove(node["name"])
            except Exception as e:
                cluster.log.error(f"Subnet connectivity test failed from '{success_node['name']}' to '{node['name']}'")
                cluster.log.verbose(f"Exception details: {e}")

    return failed_nodes


@contextmanager
def install_tcp_listener(cluster: KubernetesCluster,
                         nodes: Dict[str, NodeConfig], tcp_ports: List[str]) -> Iterator[None]:
    detect_preinstalled_python(cluster)
    nodes_without_python = {node_config['name']: node_config for host, node_config in nodes.items()
                            if cluster.context['nodes'][host]['python'] == "Not installed"}
    for node_nonfig in nodes_without_python.values():
        del nodes[node_nonfig['connect_to']]

    # currently tcp listener can be run on both python 2 and 3
    check_script = utils.read_internal('resources/scripts/simple_tcp_listener.py')
    tcp_listener = "/tmp/%s.py" % uuid.uuid4().hex
    cluster.make_group(nodes.keys()).put(io.StringIO(check_script), tcp_listener)

    skipped_nodes = {}
    nodes_to_rollback = cluster.make_group([])
    try:
        collector = CollectorCallback(cluster)
        try:
            nodes_to_rollback = group = cluster.make_group(nodes)
            with group.new_executor() as exe:
                # Run process that LISTEN TCP port
                for node in exe.group.get_ordered_members_list():
                    host = node.get_host()
                    internal_ip: str = nodes[host]['internal_address']
                    ip_version = ipaddress.ip_address(internal_ip).version
                    python_executable = cluster.context['nodes'][host]['python']['executable']
                    tcp_listener_cmd = cmd_for_ports(tcp_ports, get_start_tcp_listener_cmd(python_executable, tcp_listener, ip_version))
                    node.sudo(tcp_listener_cmd, warn=True, callback=collector)
        except GroupException as e:
            nodes_to_rollback = e.get_exited_nodes_group()
            raise

        port_in_use = re.compile(r'^(\d+) in use$')
        for host, result in collector.result.items():
            matcher = port_in_use.match(result.stderr.strip())
            if matcher is not None:
                skipped_nodes[nodes[host]["name"]] = matcher.group(1)
                del nodes[host]
            elif result.exited != 0:
                raise GroupResultException(collector.result)
            else:
                cluster.log.verbose(result)

        yield

    finally:
        with nodes_to_rollback.new_executor() as exe:
            # Kill the created during the test processes
            for node in exe.group.get_ordered_members_list():
                tcp_listener_cmd = cmd_for_ports(tcp_ports, get_stop_tcp_listener_cmd(tcp_listener))
                node.sudo(tcp_listener_cmd, warn=True)

    if skipped_nodes:
        cluster.log.warning(f"Ports in use: {skipped_nodes}")
        raise TestWarn(f"Cannot perform check on {list(skipped_nodes.keys())}: some ports are already in use. "
                       f"Use check_paas procedure if you already have installed cluster.")

    if nodes_without_python:
        cluster.log.warning(f"Nodes without python: {nodes_without_python.keys()}")
        raise TestWarn(f"Cannot perform check on {list(nodes_without_python.keys())}: python doesn't exist.")


def check_tcp_ports(cluster: KubernetesCluster) -> None:
    with TestCase(cluster, '011', 'Network', 'TCPPorts', default_results='Connected'),\
            suspend_firewalld(cluster):
        tcp_ports = ["80", "443", "179", "5473", "6443", "8443", "2379", "2380", "9091", "9093", "9094", "10250", "10254",
                     "10257", "10259", "30001", "30002"]
        nodes = {node["connect_to"]: node
                 for node in cluster.inventory['nodes']}
        host_to_ip = {host: node['internal_address'] for host, node in nodes.items()}
        with install_tcp_listener(cluster, nodes, tcp_ports):
            failed_nodes = check_tcp_connect_between_all_nodes(cluster, list(nodes.values()), tcp_ports, host_to_ip)

            if failed_nodes:
                raise TestFailure(f"Failed to connect to {len(failed_nodes)} nodes.",
                                  hint=f"Not all needed tcp ports are opened on nodes: {failed_nodes}. "
                                       f"Ports that should be opened: {tcp_ports}")


def make_reports(context: dict) -> None:
    if not context['execution_arguments'].get('disable_csv_report', False):
        context['testsuite'].save_csv(context['execution_arguments']['csv_report'], context['execution_arguments']['csv_report_delimiter'])
    if not context['execution_arguments'].get('disable_html_report', False):
        context['testsuite'].save_html(context['execution_arguments']['html_report'], context['initial_procedure'].upper())


tasks = OrderedDict({
    'ssh': {
        # todo this is useless, because flow.load_inventory already fails in case of no connectivity
        'connectivity': connection_ssh_connectivity,
        'latency': {
            'single': connection_ssh_latency_single,
            'multiple': connection_ssh_latency_multiple
        },
        'sudoer_access': connection_sudoer_access,
    },
    'network': {
        'pod_subnet_connectivity': pod_subnet_connectivity,
        'service_subnet_connectivity': service_subnet_connectivity,
        'check_tcp_ports': check_tcp_ports
    },
    'hardware': {
        'members_amount': {
            'vips': lambda cluster: hardware_members_amount(cluster, 'vip'),
            'balancers': lambda cluster: hardware_members_amount(cluster, 'balancer'),
            'control-planes': lambda cluster: hardware_members_amount(cluster, 'control-plane'),
            'workers': lambda cluster: hardware_members_amount(cluster, 'worker'),
            'total': lambda cluster: hardware_members_amount(cluster, 'all'),
        },
        'cpu': {
            'balancers': lambda cluster: hardware_cpu(cluster, 'balancer'),
            'control-planes': lambda cluster: hardware_cpu(cluster, 'control-plane'),
            'workers': lambda cluster: hardware_cpu(cluster, 'worker')
        },
        'ram': {
            'balancers': lambda cluster: hardware_ram(cluster, 'balancer'),
            'control-planes': lambda cluster: hardware_ram(cluster, 'control-plane'),
            'workers': lambda cluster: hardware_ram(cluster, 'worker')
        }
    },
    'system': {
        'distributive': system_distributive
    },
    'software': {
        'kernel': {
            'version': check_kernel_version
        },
        'thirdparties': {
            'availability': check_access_to_thirdparties
        },
        'packages': {
            'repositories': check_access_to_package_repositories,
            'availability': check_access_to_packages
        }
    }
})


class IaasAction(Action):
    def __init__(self) -> None:
        super().__init__('check iaas')

    def run(self, res: DynamicResources) -> None:
        flow.run_tasks(res, tasks)


def create_context(cli_arguments: List[str] = None) -> dict:
    cli_help = '''
    Script for checking Kubernetes cluster IAAS layer.
    
    Hot to use:

    '''

    parser = flow.new_tasks_flow_parser(cli_help, tasks=tasks)

    parser.add_argument('--csv-report',
                        default='report.csv',
                        help='define CSV report file location')

    parser.add_argument('--csv-report-delimiter',
                        default=';',
                        help='define delimiter type for CSV report')

    parser.add_argument('--html-report',
                        default='report.html',
                        help='define HTML report file location')

    parser.add_argument('--disable-csv-report',
                        action='store_true',
                        help='forcibly disable CSV report file creation')

    parser.add_argument('--disable-html-report',
                        action='store_true',
                        help='forcibly disable HTML report file creation')

    context = flow.create_context(parser, cli_arguments, procedure='check_iaas')
    context['testsuite'] = TestSuite()
    context['preserve_inventory'] = False

    return context


def main(cli_arguments: List[str] = None) -> TestSuite:
    context = create_context(cli_arguments)
    flow_ = flow.ActionsFlow([IaasAction()])
    result = flow_.run_flow(context, print_summary=False)

    context = result.context
    testsuite: TestSuite = context['testsuite']

    # Final summary should be printed only to stdout with custom formatting
    # If test results are required for parsing, they can be found in the test results files
    print(testsuite.get_final_summary())
    testsuite.print_final_status(result.logger)
    make_reports(context)
    return testsuite


if __name__ == '__main__':
    testsuite = main()
    if testsuite.is_any_test_failed():
        sys.exit(1)



#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import argparse
import sys
from typing import Callable, List, Dict

from kubemarine.core import flow, resources
from kubemarine.core.action import Action
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.group import NodeGroup
from kubemarine.core.resources import DynamicResources

HELP_DESCRIPTION = """
Script for executing shell command
    
additional arguments:
    shell_command       command to execute on nodes
"""

class CLIAction(Action):
    def __init__(self, node_group_provider: Callable[[KubernetesCluster], NodeGroup],
                 remote_args: List[str], no_stream: bool) -> None:
        super().__init__('do')
        self.node_group_provider = node_group_provider
        self.remote_args = remote_args
        self.no_stream = no_stream

    def run(self, res: DynamicResources) -> None:
        cluster = res.cluster()
        executors_group = self.node_group_provider(cluster)
        if executors_group.is_empty():
            print('Failed to find any of specified nodes or groups')
            sys.exit(1)

        result = executors_group.sudo(" ".join(self.remote_args), hide=self.no_stream, warn=True)
        if self.no_stream:
            cluster.log.debug(result)

        if result.is_any_failed():
            sys.exit(1)

        sys.exit(0)


def main(cli_arguments: List[str] = None) -> None:

    if not cli_arguments:
        cli_arguments = sys.argv[1:]

    if '--' in cli_arguments:
        kubemarine_args = cli_arguments[:cli_arguments.index('--')]
        remote_args = cli_arguments[cli_arguments.index('--') + 1:]
    else:
        kubemarine_args = cli_arguments
        remote_args = []

    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                     prog='do',
                                     description=HELP_DESCRIPTION,
                                     usage='%(prog)s [-h] [-c CONFIG] [-n NODE] [-g GROUP] [--no_stream] -- shell_command')

    parser.add_argument('-c', '--config',
                            default='cluster.yaml',
                            help='define main cluster configuration file')

    parser.add_argument('-n', '--node',
                            help='node(s) name to execute on, can be combined with groups')

    parser.add_argument('-g', '--group',
                            help='group(s) name to execute on, can be combined with nodes')

    parser.add_argument('--no_stream',
                            action='store_true',
                            help='do not stream all remote results in real-time, show node names')

    arguments = vars(parser.parse_args(kubemarine_args))
    configfile_path = arguments.get('config')

    context = flow.create_empty_context({
        'disable_dump': True,
        'log': [
            ['stdout;level=error;colorize=true;correct_newlines=true']
        ],
        'config': configfile_path,
    })
    context['preserve_inventory'] = False

    def node_group_provider(cluster: KubernetesCluster) -> NodeGroup:
        if arguments.get('node', None) is not None or arguments.get('group', None) is not None:
            executor_lists: Dict[str, List[str]] = {
                    'node': [],
                    'group': []
            }
            for executors_type in executor_lists.keys():
                executors_str = arguments.get(executors_type)
                if executors_str:
                    if "," in executors_str:
                        for executor_name in executors_str.split(','):
                            executor_lists[executors_type].append(executor_name.strip())
                    else:
                        executor_lists[executors_type].append(executors_str.strip())
            return cluster.create_group_from_groups_nodes_names(executor_lists['group'], executor_lists['node'])
        else:
            return cluster.nodes['control-plane'].get_any_member()

    no_stream: bool = arguments['no_stream']
    action = CLIAction(node_group_provider, remote_args, no_stream)
    res = resources.DynamicResources(context, silent=True)
    flow.ActionsFlow([action]).run_flow(res, print_summary=False)


if __name__ == '__main__':
    main()
